# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

AVAILABLE_METHODS = ['dbscan', 'kmeans', 'affprop', 'meanshift', 'birch', 'barber']


def get_required_params(method):
    if method == 'kmeans':
        return ['n_clusters']
    if method == 'birch':
        return ['n_clusters']



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This is a magic portal detector.
"""

################################################################################
# reuse AQ logger

import logging
from aquaduct import logger, logger_name

formatter_string = '%(name)s:%(levelname)s:[%(module)s|%(funcName)s@%(lineno)d]: %(message)s'
# create and add console handler with WARNING level to the AQ logger
formatter = logging.Formatter(formatter_string)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
ch.setLevel(logging.WARNING)  # default level is WARNING
logger.addHandler(ch)

################################################################################

if __name__ == "__main__":

    from aquaduct.utils import clui

    with clui.tictoc('Magic Portal calculations'):

        ############################################################################
        # argument parsing
        import argparse
        from aquaduct import version_nice as aquaduct_version_nice

        description_version = '''Aquaduct library version %s''' % (aquaduct_version_nice(),)
        description = '''Magic Portal (debugging tool, not for regular use)'''

        parser = argparse.ArgumentParser(description=description,
                                         formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        parser.add_argument("-t", action="store", dest="top", required=False, default=None,
                            help="Topology file.")
        parser.add_argument("-p", action="store", dest="tra", required=False, default=None,
                            help="Trajectory file(s).")
        parser.add_argument("-a", action="store", dest="areas", required=False, default=None,
                            help="Portal areas, semicolon separated.")

        args = parser.parse_args()

        areas = args.areas.split(';')

        from aquaduct.traj.sandwich import Reader, Window
        from os import pathsep

        Reader(args.top, [trj.strip() for trj in args.tra.split(pathsep)], Window(None, None, 1))  # trajectory reader

        from aquaduct.geom.convexhull import SciPyConvexHull

        print("frame", end=' ')
        for nr, area in enumerate(areas):
            print(("a%da" % (nr + 1)), ("a%dv" % (nr + 1)), end=' ')
        print("")

        for traj_reader in Reader.iterate():
            traj_reader = traj_reader.open()
            for frame in traj_reader.iterate_over_frames():
                print(frame, end=' ')
                for area in areas:
                    chull = traj_reader.parse_selection(area).chull()
                    print(chull.area, chull.volume, end=' ')
                print("")



from aquaduct.utils.helpers import listify, SmartRange
import numpy as np


################################################################################

def presort(a, b):
    s = [a[0], a[-1], b[0], b[-1]]
    return s, np.argsort(s).tolist()


def intersection_simple(a, b):
    s, sa = presort(a, b)
    ss = sorted(sa[1:3])
    if sa in [[0, 1, 2, 3], [2, 3, 0, 1], [0, 2, 1, 3], [2, 0, 3, 1], [0, 2, 3, 1], [2, 0, 1, 3]]:
        if sa[1:3] in [[1, 2], [3, 0]]:
            if s[1] == s[2]:
                return list(range(s[sa[1]], s[sa[2]] + 1))
        else:
            return list(range(s[sa[1]], s[sa[2]] + 1))
    return []


def intersection_full(a, b):
    return [aa for aa in a if aa in b]


def intersection_smartr(a, b):
    b_ = SmartRange(b)
    return [aa for aa in a if b_.isin(aa)]


def intersection_set(a, b):
    return sorted(set(a).intersection(b))


################################################################################

def glue(a, b):
    if a[-1] >= b[0]:
        return sorted(set(a + b))
    return []


def glue_simple(a, b):
    if a[-1] >= b[0]:
        if a[0] > b[-1]:
            return list(range(min(a[0], b[0]), b[-1] + 1)) + list(range(a[0], max(a[-1], b[-1]) + 1))
        return list(range(min(a[0], b[0]), max(a[-1], b[-1]) + 1))
    return []


################################################################################

@listify
def xor_full(a, b):
    ab = set(intersection_full(a, b))
    for e in glue(a, b):
        if e not in ab:
            yield e


@listify
def xor_smartr(a, b):
    ab = SmartRange(intersection_smartr(a, b))
    for e in glue(a, b):
        if not ab.isin(e):
            yield e


def xor_set(a, b):
    ab = set(intersection_set(a, b))
    return [e for e in glue(a, b) if e not in ab]


def xor_simple(a, b):
    lohi = intersection_simple(a, b)
    if len(lohi):
        lo, hi = lohi[0], lohi[-1]
        return [e for e in glue_simple(a, b) if e < lo or e > hi]
    return glue_simple(a, b)


################################################################################

def left_full(a, b):
    return intersection_full(a, xor_full(a, b))


def left_smartr(a, b):
    return intersection_smartr(a, xor_smartr(a, b))


def left_set(a, b):
    return intersection_set(a, xor_set(a, b))


def left_simple(a, b):
    # return intersection_simple(a, xor_simple(a, b))
    # intersection simple is too simple!
    return intersection_set(a, xor_simple(a, b))


################################################################################

def right_full(a, b):
    return intersection_full(b, xor_full(a, b))


def right_smartr(a, b):
    return intersection_smartr(b, xor_smartr(a, b))


def right_set(a, b):
    return intersection_set(b, xor_set(a, b))


def right_simple(a, b):
    # intersection simple is too simple!
    return intersection_set(b, xor_simple(a, b))


################################################################################

intersection = intersection_set
left = left_simple
right = right_simple



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Michał Banas
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy

class longstr(str):
    """
    Class used to specify type of default value.

    Represents Text field.
    """
    pass


class filetype(str):
    """
    Class used to specify type of default value.

    Represents Entry with file loading button.
    """
    pass


class manyfiletype(str):
    """
    Class used to specify type of default value.

    Represents Entry with file loading button which duplicate itself when previous is loaded.
    """
    pass


class dirtype(str):
    """
    Class used to specify type of default value.

    Represents Entry with dir loading button.
    """
    pass


class DefaultSection(object):
    def __init__(self, config_name, name, level, abs_level=None, additional=False):
        """
        Contrains info about section necessary to create it.

        :param config_name: Name of section in config.
        :param name: Brief label text, which will be displayed in Notebook tab.
        :param level: Entry level.
        :param abs_level: If specified section will be shown only when corresponding level was chosen.
        :param additional: If True values from that section wont be considered as default. Useful when section is mix of other sections.
        """
        self.config_name = config_name
        self.name = name
        self.level = level
        self.abs_level = abs_level
        self.additional = additional

        self.entries = []

        self._nested = False

    def add_entry(self, entry):
        """
        Allow to add new entry to the section or nest other section.
        Nested section will be showed in LabelFrame.
        :param entry: Entry or section
        :type entry DefaultEntry, DefaultSection
        """
        if isinstance(entry, DefaultEntry) or isinstance(entry, DefaultSection):
            self.entries.append(entry)
        else:
            raise TypeError("entry must be DefaultEntry or DefaultSection type.")

    def iter_entries(self):
        """
        Iterates through section entries.
        """
        def iter_inner(section):
            for entry in section.entries:
                if isinstance(entry, DefaultSection):
                    iter_inner(entry)
                else:
                    yield section, entry

        return iter_inner(self)


class DefaultEntry(object):
    def __init__(self, config_name, name, default_values, help_text, level, group_label=None, info_text=None,
                 warning_text=None, optionmenu_value=None, required=None):
        """
        Contains info about entry necessary to create it.

        :param config_name: Name of option in config.
        :param name: Brief label text, which will be displayed near widget.
        :param default_values: List of default values.
        :param help_text: Tooltip text.
        :param level: Entry level. Check LEVELS dict for adjust it.
        :param group_label: Used to group labels into frames. Content is a title of frame.
        :param info_text: If present information icon with content of that variable will be displayed.
        :param warning_text: If present warning icon with content of that variable will be displayed.
        :param optionmenu_value: OptionMenu value, which will cause display appropriate hiding frame.
        """
        self.config_name = config_name
        self.name = name

        if not isinstance(default_values, list):
            raise TypeError("Defaults values should be in the list")

        self.default_values = default_values
        self.help_text = help_text
        self.level = level
        self.group_label = group_label

        if info_text and warning_text:
            raise ValueError("There is no posibility to specify information and warning text.")

        self.info_text = info_text
        self.warning_text = warning_text
        self.optionmenu_value = optionmenu_value
        self.required = required

    @property
    def default_value(self):
        """
        Determines default value based on specified default values list.

        :return: Entry default value.
        """
        if len(self.default_values) == 1:
            if isinstance(self.default_values[0], tuple):
                default_value = self.default_values[0][0]
            elif isinstance(self.default_values[0], list):
                default_value = self.default_values[0][0]
            else:
                default_value = self.default_values[0]
        elif len(self.default_values) == 2:
            if isinstance(self.default_values[0], tuple):
                default_value = self.default_values[0][0]
            elif isinstance(self.default_values[0], list):
                default_value = self.default_values[0][0]
            # If second value is bool its checkbox.
            # If checkbox is False there is no matter what is in input widget
            elif isinstance(self.default_values[1], bool):
                default_value = self.default_values[1]
                if default_value:
                    default_value = self.default_values[0]
            else:
                default_value = self.default_values[0]

        return default_value


def is_menu(section, option):
    """
    Determines if option in section control hiding frames.

    :return: True if option control hiding frames, otherwise False.
    :rtype: bool
    """
    s = ":".join((section, option))
    for item in MENUS:
        if s == item:
            return True

    return False


def get_default_entry(section_name, option_name):
    """
    Return default entries.

    :param section_name: Name of section where option is located.
    :param option_name: Option name which default values are demanded.
    :return: DefaultEntry
    """
    for entry in get_default_section(section_name).entries:
        if option_name == entry.config_name:
            return entry

    raise RuntimeError("Entry {} in {} section does not exists.".format(option_name, section_name))


def get_default_section(section_name):
    """
    Return default section informations.

    :param section_name: Name of section which informations are demaned.
    :return: Default section informations.
    :rtype: DefaultSection
    """
    if section_name.startswith("clustering"):
        section_name = "clustering"

    if section_name.startswith("reclustering"):
        section_name = "reclustering"

    for section in DEFAULTS:
        if section.config_name == section_name and not section.additional:
            return section

    raise RuntimeError("Section {} does not exists.".format(section_name))


"""
Possible combinations:
str() -> Regular Entry
float() -> Smaller Entry
int() -> Smaller entry
bool() -> Checkbutton
filetype() -> Entry with file loading button
manyfiletype() -> Entry with file loading button with ability to duplicate itself
tuple() -> Option menu widget
any widget, bool() -> widget with checkbox
widget, float() -> create "widget_value(float_value)"

To create hiding frame set optionmenu_value to specific value and add info into MENUS list

Grouped entries will be displayed at place where first entry of that group will occur
Grouping does not work in entries with specified optionmenu_value
"""
# @formatter:off
DEFAULTS = []


###
# Initial, Easy
###
initial_section = DefaultSection(config_name="global", name="Initial", level=2, abs_level=2, additional=True)
initial_section.add_entry(DefaultEntry(config_name="top",
                                       name="Topology file: ",
                                       default_values=[manyfiletype()],
                                       help_text="Path to topology file. Aqua-Duct supports pdb, prmtop, crd, psf, ent, pqr, pdbqt, gro, top, parm7, dms, tpr, itp, mol2, data, lammpsdump, xyz, txyz, arc, gms, xml, gsd, mmtf, in topology files.",
                                       level=2,
                                       required=1))
initial_section.add_entry(DefaultEntry(config_name="trj",
                                       name="Trajectory file: ",
                                       default_values=[manyfiletype()],
                                       help_text="Path to trajectory file. Aqua-Duct supports dcd, nc, ncdf, data, lammpsdump, xtc, trr, xyz, txyz, arc, gsd, gms, log, out, trj, mdrcd, inpcrd, restrt, ent, pdbqt, pqr, gro, crd, dms, trz, mol2, config, history, mmtf, and in trajectory files.",
                                       level=2,
                                       required=1))
initial_section.add_entry(DefaultEntry(config_name="cache_dir",
                                       name="Cache directory: ",
                                       default_values=[dirtype("cache")],
                                       help_text="Allows to set path to the directory for cache data.",
                                       level=2))
initial_section.add_entry(DefaultEntry(config_name="cache_type",
                                       name="Cache type: ",
                                       default_values=[dirtype("full")],
                                       help_text="Allows to set type of coordinates cache.",
                                       level=2))

traceable_residues_nested = DefaultSection(config_name="traceable_residues", name="Traceable residues", level=2,
                                           additional=True)
traceable_residues_nested.add_entry(DefaultEntry(config_name="scope",
                                                 name="Scope: ",
                                                 default_values=["protein"],
                                                 help_text="Definition of Scope of interest.",
                                                 level=2,
                                                 required=1))
traceable_residues_nested.add_entry(DefaultEntry(config_name="object",
                                                 name="Object: ",
                                                 default_values=[str()],
                                                 help_text="Definition of Object of interest.",
                                                 level=2,
                                                 required=1))

initial_section.add_entry(traceable_residues_nested)

separate_paths_nested = DefaultSection(config_name="separate_paths", name="Separate paths", level=2, additional=True)
separate_paths_nested.add_entry(DefaultEntry(config_name="auto_barber",
                                             name="Auto Barber: ",
                                             default_values=[str()],
                                             help_text="This option allows to select molecular entity used in Auto Barber procedure. ",
                                             level=2))
separate_paths_nested.add_entry(DefaultEntry(config_name="execute",
                                              name="Execute: ",
                                              default_values=[("runonce", "run", "skip")],
                                              help_text="Option controls stage execution. It can have one of three possible values: run, runonce, and skip. If it is set to run calculations are always performed and if dump is set dump file is saved. If it is set to runonce calculations are performed if there is no dump file specified by dump option. If it is present calculations are skipped and data is loaded from the file. If it is set to skip calculations are skip and if dump is set data is loaded from the file.",
                                              level=1,
                                              info_text=" "))
initial_section.add_entry(separate_paths_nested)

DEFAULTS.append(initial_section)

###
# Clustering, Easy
###
clust_easy_section = DefaultSection(config_name="inlets_clustering", name="Clustering", level=2, abs_level=2, additional=True)
clust_easy_section.add_entry(DefaultEntry(config_name="detect_outliers",
                                          name="Detect outliers: ",
                                          default_values=[["False", "Auto"]],
                                          help_text="If set, detection of outliers is executed. It could be set as a floating point distance threshold or set to Auto.",
                                          level=1))
clust_easy_section.add_entry(DefaultEntry(config_name="singletons_outliers",
                                          name="Singletons outliers: ",
                                          default_values=["False"],
                                          help_text="Maximal size of cluster to be considered as outliers. If set to number > 0 clusters of that size are removed and their objects are moved to outliers.",
                                          level=2))
clust_easy_section.add_entry(DefaultEntry(config_name="max_level",
                                          name="Max level: ",
                                          default_values=[0],
                                          help_text="Maximal number of recursive clustering levels.",
                                          level=0,
                                          info_text=" "))

DEFAULTS.append(clust_easy_section)

###
# Global, Common
###
global_section = DefaultSection(config_name="global", name="General options", level=1)
global_section.add_entry(DefaultEntry(config_name="top",
                                      name="Topology file: ",
                                      default_values=[manyfiletype()],
                                      help_text="Path to topology file. Aqua-Duct supports pdb, prmtop, crd, psf, ent, pqr, pdbqt, gro, top, parm7, dms, tpr, itp, mol2, data, lammpsdump, xyz, txyz, arc, gms, xml, gsd, mmtf, in topology files.",
                                      level=1,
                                      required=1))
global_section.add_entry(DefaultEntry(config_name="trj",
                                      name="Trajectory file: ",
                                      default_values=[manyfiletype()],
                                      help_text="Path to trajectory file. Aqua-Duct supports dcd, nc, ncdf, data, lammpsdump, xtc, trr, xyz, txyz, arc, gsd, gms, log, out, trj, mdrcd, inpcrd, restrt, ent, pdbqt, pqr, gro, crd, dms, trz, mol2, config, history, mmtf, and in trajectory files.",
                                      level=1,
                                      required=1))
global_section.add_entry(DefaultEntry(config_name="twoway",
                                      name="Two-way scanning: ",
                                      default_values=[True],
                                      help_text="Try to use two-way scanning in the stage II.",
                                      level=0,
                                      info_text=" "))
global_section.add_entry(DefaultEntry(config_name="sandwich",
                                      name="Sandwich: ",
                                      default_values=[False],
                                      help_text="If set True trajectories are read as layers.",
                                      level=1))
global_section.add_entry(DefaultEntry(config_name="waterfall",
                                      name="Waterfall: ",
                                      default_values=[False],
                                      help_text="If set True ends of trajectories are ends of paths.",
                                      level=1))
global_section.add_entry(DefaultEntry(config_name="min_frame",
                                      name="Minimal frame: ",
                                      default_values=[0],
                                      help_text="Minimal number of frame to be read from trajectory data.",
                                      level=1))
global_section.add_entry(DefaultEntry(config_name="max_frame",
                                      name="Maximal frame: ",
                                      default_values=[str()],
                                      help_text="Maximal number of frame to be read from trajectory data. If set None trajectory is read to the last possible frame.",
                                      level=1))
global_section.add_entry(DefaultEntry(config_name="step_frame",
                                      name="Frame step: ",
                                      default_values=[1],
                                      help_text="Step used in reading trajectory. Default value of 1 stands for reading every frame. If it is greater than 1, only every step-value frame is read.",
                                      level=1))
global_section.add_entry(DefaultEntry(config_name="sps",
                                      name="Single precision storage: ",
                                      default_values=[True],
                                      help_text="Try to store data in single precission storage.",
                                      level=0))
global_section.add_entry(DefaultEntry(config_name="cache_dir",
                                      name="Cache directory: ",
                                      default_values=[dirtype("cache")],
                                      help_text="Allows to set path to the directory for cache data.",
                                      level=1))
global_section.add_entry(DefaultEntry(config_name="cache_mem",
                                      name="Memory cache: ",
                                      default_values=[False],
                                      help_text="If set True, all data will be cached in RAM.",
                                      level=1))
DEFAULTS.append(global_section)

###
# Tracking section for normal level, mixin of traceable residues and separate paths
###
tracking_section = DefaultSection(config_name="traceable_residues",
                                  name="Tracking",
                                  level=1,
                                  abs_level=1,
                                  additional=True)

traceable_residues_section = DefaultSection(config_name="traceable_residues", name="Traceable residues", level=0)
traceable_residues_section.add_entry(DefaultEntry(config_name="execute",
                                                  name="Execute: ",
                                                  default_values=[("runonce", "run", "skip")],
                                                  help_text="Option controls stage execution. It can have one of three possible values: run, runonce, and skip. If it is set to run calculations are always performed and if dump is set dump file is saved. If it is set to runonce calculations are performed if there is no dump file specified by dump option. If it is present calculations are skipped and data is loaded from the file, If it is set to skip calculations are skip and if dump is set data is loaded from the file.",
                                                  level=1,
                                                  info_text=" "))
traceable_residues_section.add_entry(DefaultEntry(config_name="dump",
                                                  name="Dump file: ",
                                                  default_values=["1_traceable_residues_data.dump"],
                                                  help_text="File name of dump data. It is used to save results of calculations or to load previously calculated data - this depends on execute option.",
                                                  level=0))
traceable_residues_section.add_entry(DefaultEntry(config_name="scope",
                                                  name="Scope: ",
                                                  default_values=[str()],
                                                  help_text="Definition of Scope of interest.",
                                                  level=1,
                                                  required=1))
traceable_residues_section.add_entry(DefaultEntry(config_name="scope_convexhull",
                                                  name="Convex hull scope: ",
                                                  default_values=[True],
                                                  help_text="Flag to set if Scope is direct or convex hull definition.",
                                                  level=1))
traceable_residues_section.add_entry(DefaultEntry(config_name="scope_everyframe",
                                                  name="Everyframe scope: ",
                                                  default_values=[False],
                                                  help_text="Flag to set Scope evaluation mode. If set True Scope is evaluated in every frame. This make sense if the definition is complex and depends on distances between molecular entities.",
                                                  level=0,
                                                  warning_text="Could be time-consuming."))
traceable_residues_section.add_entry(DefaultEntry(config_name="scope_convexhull_inflate",
                                                  name="Scope convex hull inflate: ",
                                                  default_values=[str()],
                                                  help_text="Increase (or if negative - decrease) size of the scope convex hull.",
                                                  level=0,
                                                  warning_text="Could be time-consuming."))
traceable_residues_section.add_entry(DefaultEntry(config_name="object",
                                                  name="Object: ",
                                                  default_values=[str()],
                                                  help_text="Definition of Object of interest.",
                                                  level=1,
                                                  required=1))
traceable_residues_section.add_entry(DefaultEntry(config_name="add_passing",
                                                  name="Add passing: ",
                                                  default_values=[str()],
                                                  help_text="Definition of molecules that should be added to traced molecules even if they were not present in Object.",
                                                  level=0,
                                                  warning_text="Could be time-consuming."))
DEFAULTS.append(traceable_residues_section)

tracking_section.add_entry(copy.copy(traceable_residues_section))
tracking_section.add_entry(separate_paths_nested)

tracking_analysis_nested_section = DefaultSection(config_name="analysis",
                                                  name="Calculate scope and object size",
                                                  level=1,
                                                  abs_level=1,
                                                  additional=True)
tracking_analysis_nested_section.add_entry(DefaultEntry(config_name="calculate_scope_object_size",
                                                        name="Calculate scope and object size: ",
                                                        default_values=[False],
                                                        help_text="If set to True volumes and areas of object and scope approximated by convex hulls will be calculated for each analyzed frames and saved in output CSV file.",
                                                        level=1))
tracking_analysis_nested_section.add_entry(DefaultEntry(config_name="scope_chull",
                                                        name="Scope hull definition: ",
                                                        default_values=["protein"],
                                                        help_text="Scope convex hull definition used in calculating volume and area.",
                                                        level=1))
tracking_analysis_nested_section.add_entry(DefaultEntry(config_name="object_chull",
                                                        name="Object convex hull: ",
                                                        default_values=[str()],
                                                        help_text="Object convex hull definition used in calculating volume and area.",
                                                        level=1))
tracking_analysis_nested_section.add_entry(DefaultEntry(config_name="cluster_area",
                                                        name="Cluster area: ",
                                                        default_values=[True],
                                                        help_text="If set True, clusters’ areas are estimated with kernel density estimation method (KDE).",
                                                        level=1))

tracking_section.add_entry(tracking_analysis_nested_section)
DEFAULTS.append(tracking_section)

raw_paths_section = DefaultSection(config_name="raw_paths", name="Raw paths", level=1)
raw_paths_section.add_entry(DefaultEntry(config_name="execute",
                                         name="Execute: ",
                                         default_values=[("runonce", "run", "skip")],
                                         help_text="Option controls stage execution. It can have one of three possible values: run, runonce, and skip. If it is set to run calculations are always performed and if dump is set dump file is saved. If it is set to runonce calculations are performed if there is no dump file specified by dump option. If it is present calculations are skipped and data is loaded from the file, If it is set to skip calculations are skip and if dump is set data is loaded from the file.",
                                         level=0,
                                         info_text=" "))
raw_paths_section.add_entry(DefaultEntry(config_name="dump",
                                         name="Dump file: ",
                                         default_values=["2_raw_paths_data.dump"],
                                         help_text="File name of dump data. It is used to save results of calculations or to load previously calculated data - this depends on execute option.",
                                         level=0))
raw_paths_section.add_entry(DefaultEntry(config_name="scope",
                                         name="Scope: ",
                                         default_values=[str()],
                                         help_text="Definition of Scope of interest. If None value form previous stage is used.",
                                         level=0,
                                         info_text=" "))
raw_paths_section.add_entry(DefaultEntry(config_name="scope_convexhull",
                                         name="Convex hull scope: ",
                                         default_values=[str()],
                                         help_text="Flag to set if the Scope is direct or convex hull definition.",
                                         level=0,
                                         warning_text=" "))
raw_paths_section.add_entry(DefaultEntry(config_name="scope_everyframe",
                                         name="Everyframe scope: ",
                                         default_values=[False],
                                         help_text="Flag to set Scope evaluation mode. If set True Scope is evaluated in every frame. This make sense if the definition is complex and depends on distances between molecular entities. If None value from previous stage is used.",
                                         level=0,
                                         warning_text="Could be time-consuming."))
raw_paths_section.add_entry(DefaultEntry(config_name="scope_convexhull_inflate",
                                         name="Scope convex hull inflate: ",
                                         default_values=[str()],
                                         help_text="Increase (or if negative - decrease) size of the scope convex hull. If None, value from previous stage is used.",
                                         level=0,
                                         warning_text="Could be time-consuming."))
raw_paths_section.add_entry(DefaultEntry(config_name="object",
                                         name="Object: ",
                                         default_values=[str()],
                                         help_text="Definition of Object of interest. If None value from the previous stage is used.",
                                         level=0,
                                         info_text=" "))
raw_paths_section.add_entry(DefaultEntry(config_name="clear_in_object_info",
                                         name="Clear in object info: ",
                                         default_values=[False],
                                         help_text="If it is set to True information on occupation of Object site by traceable residues calculated in the previous stage is cleared and have to be recalculated. This is useful if definition of Object was changed.",
                                         level=0,
                                         warning_text=" "))
raw_paths_section.add_entry(DefaultEntry(config_name="discard_singletons",
                                         name="Discard singletons: ",
                                         default_values=[1],
                                         help_text="If > 0, discards paths of given frame length.",
                                         level=0,
                                         info_text=" "))
raw_paths_section.add_entry(DefaultEntry(config_name="discard_empty_paths",
                                         name="Discard empty paths: ",
                                         default_values=[True],
                                         help_text="If set to True, empty paths are discarded.",
                                         level=0,
                                         info_text=" "))
DEFAULTS.append(raw_paths_section)

separate_paths_section = DefaultSection(config_name="separate_paths", name="Separate paths", level=0)
separate_paths_section.add_entry(DefaultEntry(config_name="execute",
                                              name="Execute: ",
                                              default_values=[("runonce", "run", "skip")],
                                              help_text="Option controls stage execution. It can have one of three possible values: run, runonce, and skip. If it is set to run calculations are always performed and if dump is set dump file is saved. If it is set to runonce calculations are performed if there is no dump file specified by dump option. If it is present calculations are skipped and data is loaded from the file. If it is set to skip calculations are skip and if dump is set data is loaded from the file.",
                                              level=1,
                                              info_text=" "))
separate_paths_section.add_entry(DefaultEntry(config_name="dump",
                                              name="Dump file: ",
                                              default_values=["3_separate_paths_data.dump"],
                                              help_text="File name of dump data. It is used to save results of calculations or to load previously calculated data - this depends on execute option.",
                                              level=0))
separate_paths_section.add_entry(DefaultEntry(config_name="discard_empty_paths",
                                              name="Discard empty paths: ",
                                              default_values=[True],
                                              help_text="If set to True empty paths are discarded.",
                                              level=0,
                                              info_text=" "))
separate_paths_section.add_entry(DefaultEntry(config_name="sort_by_id",
                                              name="Sort by ID: ",
                                              default_values=[True],
                                              help_text="If set to True separate paths are sorted by ID. Otherwise they are sorted in order of appearance.",
                                              level=0))
separate_paths_section.add_entry(DefaultEntry(config_name="discard_short_paths",
                                              name="Discard short paths: ",
                                              default_values=[20],
                                              help_text="This option allows to discard paths which are shorter than the threshold which is defined as total number of frames.",
                                              level=0,
                                              info_text=" "))
separate_paths_section.add_entry(DefaultEntry(config_name="discard_short_object",
                                              name="Discard short object: ",
                                              default_values=[2.0],
                                              help_text="This option allows to discard paths which objects are shorter than the threshold which is defined as total length in metric units.",
                                              level=0,
                                              warning_text=" "))
separate_paths_section.add_entry(DefaultEntry(config_name="discard_short_logic",
                                              name="Discard short logic: ",
                                              default_values=[("or", "and")],
                                              help_text="If both \"Discard short paths\" and \"Discard short object\" options are used, this option allows to set combination logic. If it is set to \"or\", a path is discarded if any of discard criterion is met. If it is set \"and\", both criteria have to be met to discard path.",
                                              level=0))
separate_paths_section.add_entry(DefaultEntry(config_name="auto_barber",
                                              name="Auto Barber: ",
                                              default_values=[str()],
                                              help_text="This option allows to select molecular entity used in Auto Barber procedure. ",
                                              level=1,
                                              group_label="Auto Barber"))
separate_paths_section.add_entry(DefaultEntry(config_name="auto_barber_mincut",
                                              name="Auto Barber mincut: ",
                                              default_values=[str()],
                                              help_text="Minimal radius of spheres used in Auto Barber. If a sphere has radius smaller then this value it is not used in AutoBarber procedure. This option can be switched off by setting it to None.",
                                              level=0,
                                              group_label="Auto Barber"))
separate_paths_section.add_entry(DefaultEntry(config_name="auto_barber_maxcut",
                                              name="Auto Barber maxcut: ",
                                              default_values=[1.2],
                                              help_text="Maximal radius of spheres used in Auto Barber. If a sphere has radius greater then this value it is not used in AutoBarber procedure. This option can be switched off by setting it to None.",
                                              level=0,
                                              group_label="Auto Barber"))
separate_paths_section.add_entry(DefaultEntry(config_name="auto_barber_mincut_level",
                                              name="Auto Barber mincut level:",
                                              default_values=[True],
                                              help_text="If set True spheres of radius smaller than mincut are resized to mincut value.",
                                              level=0,
                                              group_label="Auto Barber"))
separate_paths_section.add_entry(DefaultEntry(config_name="auto_barber_maxcut_level",
                                              name="Auto Barber maxcut level:",
                                              default_values=[True],
                                              help_text="If set True spheres of radius greater than maxcut are resized to maxcut value.",
                                              level=0,
                                              group_label="Auto Barber"))
separate_paths_section.add_entry(DefaultEntry(config_name="auto_barber_tovdw",
                                              name="Auto Barber VdW: ",
                                              default_values=[True],
                                              help_text="Correct cutting sphere by decreasing its radius by VdW radius of the closest atom.",
                                              level=0,
                                              group_label="Auto Barber"))
separate_paths_section.add_entry(DefaultEntry(config_name="allow_passing_paths",
                                              name="Allow passing paths: ",
                                              default_values=[False],
                                              help_text="If set True paths that do not enter the object are detected and added to the rest of paths as ‘passing’ paths.",
                                              level=0,
                                              warning_text="Required for hotspots and energy profiles calculations."))
separate_paths_section.add_entry(DefaultEntry(config_name="calculate_coo",
                                              name="Calculate center of object: ",
                                              default_values=[False],
                                              help_text="If set True center of object is calculated by averaging coordinates of all paths in the object area.",
                                              level=0,
                                              warning_text="Calculation of CoO may be very long for some systems."))
separate_paths_section.add_entry(DefaultEntry(config_name="separate_barber",
                                              name="Separate barber: ",
                                              default_values=[True],
                                              help_text="Apply AutoBarber for each type of traced molecules separately.",
                                              level=0))

DEFAULTS.append(separate_paths_section)

inlets_clustering_section = DefaultSection(config_name="inlets_clustering", name="Inlets clustering",
                                           level=1)
inlets_clustering_section.add_entry(DefaultEntry(config_name="execute",
                                                 name="Execute: ",
                                                 default_values=[("runonce", "run", "skip")],
                                                 help_text="Option controls stage execution. It can have one of three possible values: run, runonce, and skip. If it is set to run calculations are always performed and if dump is set dump file is saved. If it is set to runonce calculations are performed if there is no dump file specified by dump option. If it is present calculations are skipped and data is loaded from the file. If it is set to skip calculations are skip and if dump is set data is loaded from the file.",
                                                 level=1,
                                                 info_text=" "))
inlets_clustering_section.add_entry(DefaultEntry(config_name="dump",
                                                 name="Dump file: ",
                                                 default_values=["4_inlets_clustering_data.dump"],
                                                 help_text="File name of dump data. It is used to save results of calculations or to load previously calculated data - this depends on execute option.",
                                                 level=0))
inlets_clustering_section.add_entry(DefaultEntry(config_name="recluster_outliers",
                                                 name="Recluster outliers: ",
                                                 default_values=[False],
                                                 help_text="If set to True reclustering of outliers is executed according to the method defined in reclustering section.",
                                                 level=0,
                                                 info_text=" "))
inlets_clustering_section.add_entry(DefaultEntry(config_name="detect_outliers",
                                                 name="Detect outliers: ",
                                                 default_values=[["False", "Auto"]],
                                                 help_text="If set, detection of outliers is executed. It could be set as a floating point distance threshold or set to Auto.",
                                                 level=1))
inlets_clustering_section.add_entry(DefaultEntry(config_name="singletons_outliers",
                                                 name="Singletons outliers: ",
                                                 default_values=["False"],
                                                 help_text="Maximal size of cluster to be considered as outliers. If set to number > 0 clusters of that size are removed and their objects are moved to outliers.",
                                                 level=1))
inlets_clustering_section.add_entry(DefaultEntry(config_name="max_level",
                                                 name="Max level: ",
                                                 default_values=[0],
                                                 help_text="Maximal number of recursive clustering levels.",
                                                 level=0,
                                                 info_text=" "))
inlets_clustering_section.add_entry(DefaultEntry(config_name="create_master_paths",
                                                 name="Create master paths: ",
                                                 default_values=[True],
                                                 help_text="If set to True master paths are created (fast CPU and big RAM recommended; 50k frames long simulation may need ca 20GB of memory)",
                                                 level=1,
                                                 warning_text="Required for energy profiles calculations and could be time-consuming"))
inlets_clustering_section.add_entry(DefaultEntry(config_name="master_paths_amount",
                                                 name="Master paths amount",
                                                 default_values=[str()],
                                                 help_text="Allows to limit number of single paths used for master paths calculations. If it is a number in range ``(0,1)``, then it is interpreted as a percent number of paths to be used. It is is a integer number ``>= 1`` it is an absoulte number of used paths.",
                                                 level=0))
inlets_clustering_section.add_entry(DefaultEntry(config_name="separate_master",
                                                 name="Separate master paths",
                                                 default_values=[False],
                                                 help_text="If set to True, master paths are generated for each of traced molecules' types separately.",
                                                 level=0))
inlets_clustering_section.add_entry(DefaultEntry(config_name="separate_master_all",
                                                 name="Separate master paths all",
                                                 default_values=[True],
                                                 help_text="If separate_master is used and this option is set True, master paths will be also generated for all traced molecules",
                                                 level=0))
inlets_clustering_section.add_entry(DefaultEntry(config_name="exclude_passing_in_clustering",
                                                 name="Exclude passing in clustering: ",
                                                 default_values=[True],
                                                 help_text="If set to True passing paths are not clustered with normal paths.",
                                                 level=0,
                                                 group_label="Passing paths",
                                                 info_text=" "))
inlets_clustering_section.add_entry(DefaultEntry(config_name="add_passing_to_clusters",
                                                 name="Add passing to clusters: ",
                                                 default_values=[str()],
                                                 help_text="Allows to run procedure for adding passing paths inlets to clusters with Auto Barber method. To enable this the option should be set to molecular entity that will be used by Auto Barber.",
                                                 level=0,
                                                 group_label="Passing paths",
                                                 info_text=" "))
inlets_clustering_section.add_entry(DefaultEntry(config_name="join_clusters",
                                                 name="Join clusters: ",
                                                 default_values=[str()],
                                                 help_text="This option allows to join selected clusters. Clusters’ IDs joined with + character lists clusters to be joined together. Several such blocks separated by space can be used. For example, if set to 1+3+4 5+6 clusters 1, 3, and 4 will be joined in one cluster and cluster 5, and 6 will be also joined in another cluster.",
                                                 level=1,
                                                 group_label="Post Clustering Optimization"))
inlets_clustering_section.add_entry(DefaultEntry(config_name="renumber_clusters",
                                                 name="Renumber clusters: ",
                                                 default_values=[False],
                                                 help_text="If set True, clusters have consecutive numbers starting from 1 (or 0 if outliers are present) starting from the bigest cluster.",
                                                 level=1,
                                                 group_label="Post Clustering Optimization"))
inlets_clustering_section.add_entry(DefaultEntry(config_name="inlets_center",
                                                 name="Central point for inlets: ",
                                                 default_values=[['cos','coo']],
                                                 help_text="Central point for inlets influencs cluster area calculations and cluster countours.",
                                                 warning_text="If `coo` is selected `calculate_coo` should be set `True`. ",
                                                 level=1))
inlets_clustering_section.add_entry(DefaultEntry(config_name="remove_inlets",
                                                 name="Remove inlets from specified clusters: ",
                                                 default_values=[""],
                                                 help_text="Inlets in specified clusters are removed - corresponding paths are changed.",
                                                 warning_text="",
                                                 level=1))
inlets_clustering_section.add_entry(DefaultEntry(config_name="clustering_order",
                                                 name="Order of clustering operations: ",
                                                 default_values=[['old-school','aquarius']],
                                                 help_text="Order in which clustering related operations are executed.",
                                                 warning_text="",
                                                 level=1))



DEFAULTS.append(inlets_clustering_section)

clustering_section = DefaultSection(config_name="clustering", name="Clustering", level=1)
clustering_section.add_entry(DefaultEntry(config_name="name",
                                          name="Name: ",
                                          default_values=["clustering"],
                                          help_text="Used to refer other clustering method in \"Recursive clustering\" option",
                                          level=2))
clustering_section.add_entry(DefaultEntry(config_name="method",
                                          name="Method: ",
                                          default_values=[
                                              ("barber", "dbscan", "affprop", "meanshift", "birch", "kmeans")],
                                          help_text="Name of clustering method. ",
                                          level=2))

# Barber options
clustering_section.add_entry(DefaultEntry(config_name="auto_barber",
                                          name="Auto barber: ",
                                          default_values=[str()],
                                          help_text="This option allows to select molecular entity used in Auto Barber procedure.",
                                          level=2,
                                          optionmenu_value="barber"))
clustering_section.add_entry(DefaultEntry(config_name="auto_barber_mincut",
                                          name="Auto Barber mincut: ",
                                          default_values=[str()],
                                          help_text="Minimal radius of spheres used in Auto Barber. If a sphere has radius smaller than this value, it is not used to cut. This option can be switched off by setting it to None.",
                                          level=0,
                                          optionmenu_value="barber"))
clustering_section.add_entry(DefaultEntry(config_name="auto_barber_maxcut",
                                          name="Auto Barber maxcut: ",
                                          default_values=[2.8],
                                          help_text="Maximal radius of spheres used in Auto Barber. If a sphere has radius greater than this value, it is not used to cut. This option can be switched off by setting it to None.",
                                          level=0,
                                          optionmenu_value="barber"))
clustering_section.add_entry(DefaultEntry(config_name="auto_barber_mincut_level",
                                          name="Auto Barber mincut level: ",
                                          default_values=[True],
                                          help_text="If set True, spheres of radius less than mincut are resized to mincut value.",
                                          level=0,
                                          optionmenu_value="barber"))
clustering_section.add_entry(DefaultEntry(config_name="auto_barber_maxcut_level",
                                          name="Auto Barber maxcut level:",
                                          default_values=[True],
                                          help_text="If set True, spheres of radius greater than maxcut are resized to maxcut value.",
                                          level=0,
                                          optionmenu_value="barber"))
clustering_section.add_entry(DefaultEntry(config_name="auto_barber_tovdw",
                                          name="Auto Barber VdW: ",
                                          default_values=[True],
                                          help_text="If set True, cutting of spheres is corrected by decreasing its radius by VdW radius of the closest atom.",
                                          level=0,
                                          optionmenu_value="barber"))

# Dbscan options
clustering_section.add_entry(DefaultEntry(config_name="eps",
                                          name="Maximum distance: ",
                                          default_values=[float()],
                                          help_text="The maximum distance between two samples for them to be considered as in the same neighborhood.",
                                          level=0,
                                          optionmenu_value="dbscan"))
clustering_section.add_entry(DefaultEntry(config_name="min_samples",
                                          name="Maxium samples: ",
                                          default_values=[int()],
                                          help_text="The number of samples (or total weight) in a neighborhood for a point to be considered as a core point. This includes the point itself.",
                                          level=0, optionmenu_value="dbscan"))
clustering_section.add_entry(DefaultEntry(config_name="metric",
                                          name="Metric: ",
                                          default_values=[("euclidean", "cityblock", "cosine", "manhattan")],
                                          help_text="The metric to use when calculating distance between instances in a feature array.",
                                          level=0,
                                          optionmenu_value="dbscan"))
clustering_section.add_entry(DefaultEntry(config_name="algorithm",
                                          name="Algorithm: ",
                                          default_values=[("auto", "ball_tree", "kd_tree", "brute")],
                                          help_text="The algorithm to be used by the NearestNeighbors module to compute pointwise distances and find nearest neighbors.",
                                          level=0,
                                          optionmenu_value="dbscan"))
clustering_section.add_entry(DefaultEntry(config_name="leaf_size",
                                          name="Leaf size: ",
                                          default_values=[int()],
                                          help_text="Leaf size passed to BallTree or cKDTree.",
                                          level=0,
                                          optionmenu_value="dbscan"))

# Affprop options
clustering_section.add_entry(DefaultEntry(config_name="damping",
                                          name="Damping factor: ",
                                          default_values=[float()],
                                          help_text="Damping factor between 0.5 and 1.",
                                          level=0,
                                          optionmenu_value="affprop"))
clustering_section.add_entry(DefaultEntry(config_name="convergence_iter",
                                          name="Maximum no effect iterations: ",
                                          default_values=[int()],
                                          help_text="Number of iterations with no change in the number of estimated clusters that stops the convergence.",
                                          level=0,
                                          optionmenu_value="affprop"))
clustering_section.add_entry(DefaultEntry(config_name="max_iter",
                                          name="Maximum number of iterations: ",
                                          default_values=[int()],
                                          help_text="Maximum number of iterations.",
                                          level=0,
                                          optionmenu_value="affprop"))
clustering_section.add_entry(DefaultEntry(config_name="preference",
                                          name="Preference: ",
                                          default_values=[float()],
                                          help_text="Points with larger values of preferences are more likely to be chosen as exemplars.",
                                          level=0,
                                          optionmenu_value="affprop"))

# Meanshift options
clustering_section.add_entry(DefaultEntry(config_name="bandwidth",
                                          name="Bandwidth: ",
                                          default_values=["Auto"],
                                          help_text="Bandwidth used in the RBF kernel. If Auto or None automatic method for bandwidth estimation is used.",
                                          level=2,
                                          optionmenu_value="meanshift"))
clustering_section.add_entry(DefaultEntry(config_name="cluster_all",
                                          name="Cluster all points: ",
                                          default_values=[bool()],
                                          help_text="If true, then all points are clustered, even those orphans that are not within any kernel.",
                                          level=0,
                                          optionmenu_value="meanshift"))
clustering_section.add_entry(DefaultEntry(config_name="bin_seeding",
                                          name="bin_seeding",
                                          default_values=[bool()],
                                          help_text="If true, initial kernel locations are not locations of all points, but rather the location of the discretized version of points, where points are binned onto a grid whose coarseness corresponds to the bandwidth.",
                                          level=0,
                                          optionmenu_value="meanshift"))
clustering_section.add_entry(DefaultEntry(config_name="min_bin_freq",
                                          name="min_bin_freq",
                                          default_values=[int()],
                                          help_text="To speed up the algorithm, accept only those bins with at least min_bin_freq points as seeds. If not defined, set to 1.",
                                          level=0,
                                          optionmenu_value="meanshift"))

# Birch options
clustering_section.add_entry(DefaultEntry(config_name="threshold",
                                          name="Threshold: ",
                                          default_values=[float()],
                                          help_text="The radius of the subcluster obtained by merging a new sample and the closest subcluster should be smaller than the threshold. Otherwise a new subcluster is started.",
                                          level=0,
                                          optionmenu_value="birch"))
clustering_section.add_entry(DefaultEntry(config_name="branching_factor",
                                          name="Branching factor: ",
                                          default_values=[int()],
                                          help_text="Maximum number of CF subclusters in each node.",
                                          level=0,
                                          optionmenu_value="birch"))
clustering_section.add_entry(DefaultEntry(config_name="n_clusters",
                                          name="Cluster number: ",
                                          default_values=[int()],
                                          help_text="Number of clusters after the final clustering step, which treats the subclusters from the leaves as new samples. By default, this final clustering step is not performed and the subclusters are returned as they are.",
                                          level=2,
                                          optionmenu_value="birch"))

# Kmeans options
clustering_section.add_entry(DefaultEntry(config_name="n_clusters",
                                          name="Cluster number: ",
                                          default_values=[int()],
                                          help_text="The number of clusters to form as well as the number of centroids to generate.",
                                          level=2,
                                          optionmenu_value="kmeans"))
clustering_section.add_entry(DefaultEntry(config_name="max_iter",
                                          name="Maximum number of iterations: ",
                                          default_values=[int()],
                                          help_text="Maximum number of iterations of the k-means algorithm for a single run.",
                                          level=0,
                                          optionmenu_value="kmeans"))
clustering_section.add_entry(DefaultEntry(config_name="n_init",
                                          name="n_init",
                                          default_values=[int()],
                                          help_text="Number of times the k-means algorithm will be run with different centroid seeds. The final results will be the best output of n_init consecutive runs in terms of inertia.",
                                          level=0,
                                          optionmenu_value="kmeans"))
clustering_section.add_entry(DefaultEntry(config_name="init",
                                          name="Init method: ",
                                          default_values=[("k-means++", "random")],
                                          help_text="Method for initialization, defaults to k-means++. Can be one of following: k-means++ or random.",
                                          level=0,
                                          optionmenu_value="kmeans"))
clustering_section.add_entry(DefaultEntry(config_name="tol",
                                          name="Tolerance: ",
                                          default_values=[float()],
                                          help_text="Relative tolerance with regards to inertia to declare convergence.",
                                          level=0,
                                          optionmenu_value="kmeans"))

clustering_section.add_entry(DefaultEntry(config_name="recursive_clustering",
                                          name="Recursive clustering: ",
                                          default_values=["clustering"],
                                          help_text="If it is set to name of some section that holds clustering method settings this method will be called in the next recursion of clusteriation. Default value for reclustering is None.",
                                          level=1))
clustering_section.add_entry(DefaultEntry(config_name="recursive_threshold",
                                          name="Recursive threshold: ",
                                          default_values=[str()],
                                          help_text="Allows to set threshold that excludes clusters of certain size from reclustering. Value of this option comprises of operator and value. Operator can be one of the following: >, >=, <=, <. Value have to be expressed as floating number and it have to be in the range of 0 to 1. One can use several definitions separated by a space character. Only clusters of size complying with all thresholds definitions are submitted to reclustering.",
                                          level=1))


DEFAULTS.append(clustering_section)

reclustering_section = DefaultSection(config_name="reclustering", name="Reclustering", level=0)
reclustering_section.add_entry(DefaultEntry(config_name="name",
                                            name="Name: ",
                                            default_values=["reclustering"],
                                            help_text="Used to refer other clustering method in \"Recursive clustering\" option",
                                            level=1))
reclustering_section.add_entry(DefaultEntry(config_name="method",
                                            name="Method: ",
                                            default_values=[
                                                ("barber", "dbscan", "affprop", "meanshift", "birch", "kmeans")],
                                            help_text="Name of clustering method. It has to be one of the following: barber, dbscan, affprop, meanshift, birch, kmeans. Default value depends whether it is clustering section (barber) or reclustering section (dbscan).",
                                            level=1))

# Barber options
reclustering_section.add_entry(DefaultEntry(config_name="auto_barber",
                                            name="Auto barber: ",
                                            default_values=[str()],
                                            help_text="This option allows to select molecular entity used in Auto Barber procedure.",
                                            level=1,
                                            optionmenu_value="barber"))
reclustering_section.add_entry(DefaultEntry(config_name="auto_barber_mincut",
                                            name="Auto Barber mincut: ",
                                            default_values=[str()],
                                            help_text="Minimal radius of spheres used in Auto Barber. If a sphere has radius smaller than this value, it is not used to cut. This option can be switched off by setting it to None.",
                                            level=0,
                                            optionmenu_value="barber"))
reclustering_section.add_entry(DefaultEntry(config_name="auto_barber_maxcut",
                                            name="Auto Barber maxcut: ",
                                            default_values=[str()],
                                            help_text="Maximal radius of spheres used in Auto Barber. If a sphere has radius greater than this value, it is not used to cut. This option can be switched off by setting it to None.",
                                            level=0,
                                            optionmenu_value="barber"))
reclustering_section.add_entry(DefaultEntry(config_name="auto_barber_mincut_level",
                                            name="Auto Barber mincut level: ",
                                            default_values=[bool()],
                                            help_text="If set True, spheres of radius less than mincut are resized to mincut value.",
                                            level=0,
                                            optionmenu_value="barber"))
reclustering_section.add_entry(DefaultEntry(config_name="auto_barber_maxcut_level",
                                            name="Auto Barber maxcut level:",
                                            default_values=[bool()],
                                            help_text="If set True, spheres of radius greater than maxcut are resized to maxcut value.",
                                            level=0,
                                            optionmenu_value="barber"))
reclustering_section.add_entry(DefaultEntry(config_name="auto_barber_tovdw",
                                            name="Auto Barber VdW: ",
                                            default_values=[bool()],
                                            help_text="If set True, cutting of spheres is corrected by decreasing its radius by VdW radius of the closest atom.",
                                            level=0,
                                            optionmenu_value="barber"))

# Dbscan options
reclustering_section.add_entry(DefaultEntry(config_name="eps",
                                            name="Maximum distance: ",
                                            default_values=[float()],
                                            help_text="The maximum distance between two samples for them to be considered as in the same neighborhood.",
                                            level=0,
                                            optionmenu_value="dbscan"))
reclustering_section.add_entry(DefaultEntry(config_name="min_samples",
                                            name="Maximum samples: ",
                                            default_values=[int()],
                                            help_text="The number of samples (or total weight) in a neighborhood for a point to be considered as a core point. This includes the point itself.",
                                            level=0, optionmenu_value="dbscan"))
reclustering_section.add_entry(DefaultEntry(config_name="metric",
                                            name="Metric: ",
                                            default_values=[("euclidean", "cityblock", "cosine", "manhattan")],
                                            help_text="The metric to use when calculating distance between instances in a feature array.",
                                            level=0,
                                            optionmenu_value="dbscan"))
reclustering_section.add_entry(DefaultEntry(config_name="algorithm",
                                            name="Algorithm: ",
                                            default_values=[("auto", "ball_tree", "kd_tree", "brute")],
                                            help_text="The algorithm to be used by the NearestNeighbors module to compute pointwise distances and find nearest neighbors.",
                                            level=0,
                                            optionmenu_value="dbscan"))
reclustering_section.add_entry(DefaultEntry(config_name="leaf_size",
                                            name="Leaf size: ",
                                            default_values=[int()],
                                            help_text="Leaf size passed to BallTree or cKDTree.",
                                            level=0,
                                            optionmenu_value="dbscan"))

# Affprop options
reclustering_section.add_entry(DefaultEntry(config_name="damping",
                                            name="Damping factor: ",
                                            default_values=[float()],
                                            help_text="Damping factor between 0.5 and 1.",
                                            level=0,
                                            optionmenu_value="affprop"))
reclustering_section.add_entry(DefaultEntry(config_name="convergence_iter",
                                            name="Maximum no effect iterations: ",
                                            default_values=[int()],
                                            help_text="Number of iterations with no change in the number of estimated clusters that stops the convergence.",
                                            level=0,
                                            optionmenu_value="affprop"))
reclustering_section.add_entry(DefaultEntry(config_name="max_iter",
                                            name="Maximum number of iterations: ",
                                            default_values=[int()],
                                            help_text="Maximum number of iterations.",
                                            level=0,
                                            optionmenu_value="affprop"))
reclustering_section.add_entry(DefaultEntry(config_name="preference",
                                            name="Preference: ",
                                            default_values=[float()],
                                            help_text="Points with larger values of preferences are more likely to be chosen as exemplars.",
                                            level=0,
                                            optionmenu_value="affprop"))

# Meanshift options
reclustering_section.add_entry(DefaultEntry(config_name="bandwidth",
                                            name="Bandwidth: ",
                                            default_values=["Auto"],
                                            help_text="Bandwidth used in the RBF kernel. If Auto or None automatic method for bandwidth estimation is used.",
                                            level=1,
                                            optionmenu_value="meanshift"))
reclustering_section.add_entry(DefaultEntry(config_name="cluster_all",
                                            name="Cluster all points: ",
                                            default_values=[bool()],
                                            help_text="If true, then all points are clustered, even those orphans that are not within any kernel.",
                                            level=0,
                                            optionmenu_value="meanshift"))
reclustering_section.add_entry(DefaultEntry(config_name="bin_seeding",
                                            name="Bin seeding: ",
                                            default_values=[bool()],
                                            help_text="If true, initial kernel locations are not locations of all points, but rather the location of the discretized version of points, where points are binned onto a grid whose coarseness corresponds to the bandwidth.",
                                            level=0,
                                            optionmenu_value="meanshift"))
reclustering_section.add_entry(DefaultEntry(config_name="min_bin_freq",
                                            name="Minimum bin frequency",
                                            default_values=[int()],
                                            help_text="To speed up the algorithm, accept only those bins with at least min_bin_freq points as seeds. If not defined, set to 1.",
                                            level=0,
                                            optionmenu_value="meanshift"))

# Birch options
reclustering_section.add_entry(DefaultEntry(config_name="threshold",
                                            name="Threshold: ",
                                            default_values=[float()],
                                            help_text="The radius of the subcluster obtained by merging a new sample and the closest subcluster should be smaller than the threshold. Otherwise a new subcluster is started.",
                                            level=0,
                                            optionmenu_value="birch"))
reclustering_section.add_entry(DefaultEntry(config_name="branching_factor",
                                            name="Branching factor: ",
                                            default_values=[int()],
                                            help_text="Maximum number of CF subclusters in each node.",
                                            level=0,
                                            optionmenu_value="birch"))
reclustering_section.add_entry(DefaultEntry(config_name="n_clusters",
                                            name="Cluster number: ",
                                            default_values=[int()],
                                            help_text="Number of clusters after the final clustering step, which treats the subclusters from the leaves as new samples. By default, this final clustering step is not performed and the subclusters are returned as they are.",
                                            level=1,
                                            optionmenu_value="birch"))

# Kmeans options
reclustering_section.add_entry(DefaultEntry(config_name="n_clusters",
                                            name="Cluster number: ",
                                            default_values=[int()],
                                            help_text="The number of clusters to form as well as the number of centroids to generate.",
                                            level=1,
                                            optionmenu_value="kmeans"))
reclustering_section.add_entry(DefaultEntry(config_name="max_iter",
                                            name="Maximum number of iterations: ",
                                            default_values=[int()],
                                            help_text="Maximum number of iterations of the k-means algorithm for a single run.",
                                            level=0,
                                            optionmenu_value="kmeans"))
reclustering_section.add_entry(DefaultEntry(config_name="n_init",
                                            name="N Init: ",
                                            default_values=[int()],
                                            help_text="Number of times the k-means algorithm will be run with different centroid seeds. The final results will be the best output of n_init consecutive runs in terms of inertia.",
                                            level=0,
                                            optionmenu_value="kmeans"))
reclustering_section.add_entry(DefaultEntry(config_name="init",
                                            name="Init method: ",
                                            default_values=[("k-means++", "random")],
                                            help_text="Method for initialization, defaults to k-means++. Can be one of following: k-means++ or random.",
                                            level=0,
                                            optionmenu_value="kmeans"))
reclustering_section.add_entry(DefaultEntry(config_name="tol",
                                            name="Tolerance: ",
                                            default_values=[float()],
                                            help_text="Relative tolerance with regards to inertia to declare convergence.",
                                            level=0,
                                            optionmenu_value="kmeans"))

reclustering_section.add_entry(DefaultEntry(config_name="recursive_clustering",
                                            name="Recursive clustering: ",
                                            default_values=[str()],
                                            help_text="If it is set to name of some section that holds clustering method settings this method will be called in the next recursion of clusteriation. Default value for reclustering is None.",
                                            level=1))
reclustering_section.add_entry(DefaultEntry(config_name="recursive_threshold",
                                            name="Recursive threshold: ",
                                            default_values=[str()],
                                            help_text="Allows to set threshold that excludes clusters of certain size from reclustering. Value of this option comprises of operator and value. Operator can be one of the following: >, >=, <=, <. Value have to be expressed as floating number and it have to be in the range of 0 to 1. One can use several definitions separated by a space character. Only clusters of size complying with all thresholds definitions are submitted to reclustering.",
                                            level=1))
DEFAULTS.append(reclustering_section)

analysis_section = DefaultSection(config_name="analysis", name="Analysis", level=0)
analysis_section.add_entry(DefaultEntry(config_name="execute",
                                        name="Execute: ",
                                        default_values=[("run", "runonce", "skip")],
                                        help_text="Option controls stage execution. It can have one of three possible values: run, runonce, and skip. If it is set to run or runonce stage is executed and results is saved according to save option. If it is set to skip stage is skipped.",
                                        level=1))
analysis_section.add_entry(DefaultEntry(config_name="save",
                                        name="Save file: ",
                                        default_values=["5_analysis_results.txt"],
                                        help_text="File name for saving results.",
                                        level=0))
analysis_section.add_entry(DefaultEntry(config_name="dump_config",
                                        name="Dump config: ",
                                        default_values=[True],
                                        help_text="If set to True configuration options, as seen by Valve, are added to the head of results.",
                                        level=0))
analysis_section.add_entry(DefaultEntry(config_name="calculate_scope_object_size",
                                        name="Calculate scope and object size: ",
                                        default_values=[False],
                                        help_text="If set to True volumes and areas of object and scope approximated by convex hulls will be calculated for each analyzed frames and saved in output CSV file.",
                                        level=0))
analysis_section.add_entry(DefaultEntry(config_name="scope_chull",
                                        name="Scope hull definition: ",
                                        default_values=[str()],
                                        help_text="Scope convex hull definition used in calculating volume and area.",
                                        level=0))
analysis_section.add_entry(DefaultEntry(config_name="scope_chull_inflate",
                                        name="Scope convexhull inflate: ",
                                        default_values=[str()],
                                        help_text="Increase (or if negative - decrease) size of the scope convex hull.",
                                        level=0))
analysis_section.add_entry(DefaultEntry(config_name="object_chull",
                                        name="Object convex hull: ",
                                        default_values=[str()],
                                        help_text="Object convex hull definition used in calculating volume and area.",
                                        level=0))
analysis_section.add_entry(DefaultEntry(config_name="cluster_area",
                                        name="Cluster area: ",
                                        default_values=[True],
                                        help_text="If set True, clusters’ areas are estimated with kernel density estimation method (KDE).",
                                        level=1))

DEFAULTS.append(analysis_section)

visualize_section = DefaultSection(config_name="visualize", name="Visualize", level=2)
visualize_section.add_entry(DefaultEntry(config_name="execute",
                                         name="Execute: ",
                                         default_values=[("run", "runonce", "skip")],
                                         help_text="Option controls stage execution. It can have one of three possible values: run, runonce, and skip. If it is set to run or runonce stage is executed and results is saved according to save option. If it is set to skip stage is skipped.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="save",
                                         name="Save file: ",
                                         default_values=["6_visualize_results.py"],
                                         help_text="File name for saving results.",
                                         level=0))
visualize_section.add_entry(DefaultEntry(config_name="all_paths_raw",
                                         name="All paths raw: ",
                                         default_values=[True],
                                         help_text="If True produces one object in PyMOL that holds all paths visualized by raw coordinates.",
                                         level=2,
                                         group_label="Raw paths",
                                         warning_text=" "))
visualize_section.add_entry(DefaultEntry(config_name="all_paths_smooth",
                                         name="All paths smooth: ",
                                         default_values=[True],
                                         help_text="If True produces one object in PyMOL that holds all paths visualized by smooth coordinates.",
                                         level=2,
                                         group_label="Smooth paths"))
visualize_section.add_entry(DefaultEntry(config_name="all_paths_split",
                                         name="All split paths: ",
                                         default_values=[True],
                                         help_text="If is set True objects produced by all_paths_raw and all_paths_smooth are split into Incoming, Object, and Outgoing parts and visualized as three different objects.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="all_paths_raw_io",
                                         name="All raw paths io: ",
                                         default_values=[True],
                                         help_text="If set True arrows pointing beginning and end of paths are displayed oriented accordingly to raw paths orientation.",
                                         level=1,
                                         group_label="Raw paths",
                                         warning_text=" "))
visualize_section.add_entry(DefaultEntry(config_name="all_paths_smooth_io",
                                         name="All paths smooth io: ",
                                         default_values=[True],
                                         help_text="If set True arrows pointing beginning and end of paths are displayed oriented accordingly to smooth paths orientation.",
                                         level=1,
                                         group_label="Smooth paths"))
visualize_section.add_entry(DefaultEntry(config_name="all_paths_amount",
                                         name="Paths limit: ",
                                         default_values=[str()],
                                         help_text="Allows to limit number of visualised paths. If it is a number in range (0,1), then it is interpreted as a percent number of paths to be visualized. It is is a integer number >= 1 it is total number of all_paths visualized.",
                                         level=0))
visualize_section.add_entry(DefaultEntry(config_name="simply_smooths",
                                         name="Smoothing simplification: ",
                                         default_values=[("RecursiveVector", "HobbitVector", "OneWayVector",
                                                          "RecursiveTriangle", "HobbitTriangle", "OneWayTriangle"),
                                                         float()],
                                         help_text="Option indicates linear simplification method to be used in plotting smooth paths. Simplification removes points which do not (or almost do not) change the shape of smooth path. Optionally name of the method can be followed by a threshold value. For sane values of thresholds see appropriate documentation of each method. Default values work well. This option is not case sensitive. It is recommended to use default method or HobbitVector method.",
                                         level=0))
visualize_section.add_entry(DefaultEntry(config_name="paths_raw",
                                         name="Paths raw: ",
                                         default_values=[True],
                                         help_text="If set True raw paths are displayed as separate objects or as one object with states corresponding to number of path.",
                                         level=2,
                                         group_label="Raw paths",
                                         warning_text=" "))
visualize_section.add_entry(DefaultEntry(config_name="paths_smooth",
                                         name="Paths smooth: ",
                                         default_values=[True],
                                         help_text="If set True smooth paths are displayed as separate objects or as one object with states corresponding to number of path.",
                                         level=2,
                                         group_label="Smooth paths"))
visualize_section.add_entry(DefaultEntry(config_name="paths_raw_io",
                                         name="Paths raw io: ",
                                         default_values=[True],
                                         help_text="If set True arrows indicating beginning and end of paths, oriented accordingly to raw paths, are displayed as separate objects or as one object with states corresponding to number of paths.",
                                         level=1,
                                         group_label="Raw paths",
                                         warning_text=" "))
visualize_section.add_entry(DefaultEntry(config_name="paths_smooth_io",
                                         name="Paths smooth io: ",
                                         default_values=[True],
                                         help_text="If set True arrows indicating beginning and end of paths, oriented accordingly to smooth paths, are displayed as separate objects or as one object with states corresponding to number of paths.",
                                         level=1,
                                         group_label="Smooth paths"))
visualize_section.add_entry(DefaultEntry(config_name="paths_states",
                                         name="Paths states: ",
                                         default_values=[True],
                                         help_text="If True objects displayed by paths_raw, paths_smooth, paths_raw_io, and paths_smooth_io are displayed as one object with states corresponding to number of paths. Otherwise they are displayed as separate objects.",
                                         level=2))
visualize_section.add_entry(DefaultEntry(config_name="ctypes_raw",
                                         name="Ctypes raw: ",
                                         default_values=[True],
                                         help_text="Displays raw paths in a similar manner as non split all_paths_raw but each cluster type is displayed in separate object.",
                                         level=2,
                                         group_label="Raw paths"))
visualize_section.add_entry(DefaultEntry(config_name="ctypes_smooth",
                                         name="Ctypes smooth: ",
                                         default_values=[True],
                                         help_text="Displays smooth paths in a similar manner as non split all_paths_smooth but each cluster type is displayed in separate object.",
                                         level=2,
                                         group_label="Smooth paths"))
visualize_section.add_entry(DefaultEntry(config_name="ctypes_amount",
                                         name="Ctypes limit: ",
                                         default_values=[float()],
                                         help_text="Allows to limit number of visualised ctypes. If it is a number in range (0,1), then it is interpreted as percent number of ctypes to be visualized. It is is a integer number >= 1, it is total number of visualized ctypes.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="split_by_type",
                                         name="Split by type: ",
                                         default_values=[False],
                                         help_text="If True, visualizations of all objects that correspond to particular types of traced molecules are splited. Appropriate molecule name is added to created objects.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="retain_all_types",
                                         name="Retain all types: ",
                                         default_values=[False],
                                         help_text="If True and *split_by_type* option is True as well, standard visualization for all types of traced moelcules will be generated as well.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="inlets_clusters",
                                         name="Visualize cluster of inlets: ",
                                         default_values=[True],
                                         help_text="If set True, clusters of inlets are visualized.",
                                         level=2))
visualize_section.add_entry(DefaultEntry(config_name="inlets_clusters_amount",
                                         name="Inlets limit: ",
                                         default_values=[str()],
                                         help_text="Allows to limit number of visualised inlets. If it is a number in range (0,1) then it is interpreted as percent number of inlets to be visualized. It is is a integer number >= 1 it is total number of visualized inlets.",
                                         level=0,
                                         warning_text=" "))
visualize_section.add_entry(DefaultEntry(config_name="show_molecule",
                                         name="Show molecule: ",
                                         default_values=[["protein", "False"]],
                                         help_text="If is set to selection of some molecular object in the system, for example to protein, this object is displayed.",
                                         level=2))
visualize_section.add_entry(DefaultEntry(config_name="show_molecule_frames",
                                         name="Molecule frames: ",
                                         default_values=[0],
                                         help_text="Allows to indicate which frames of object defined by show_molecule should be displayed. It is possible to set several frames. In that case frames would be displayed as states.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="show_scope_chull",
                                         name="Show scope convex hull: ",
                                         default_values=[["protein", str()]],
                                         help_text="If is set to selection of some molecular object in the system, for example to protein, convex hull of this object is displayed.",
                                         level=2))
visualize_section.add_entry(DefaultEntry(config_name="show_scope_chull_inflate",
                                         name="Show scope convex hull inflate: ",
                                         default_values=[str()],
                                         help_text="Increase (or if negative decrease) size of the scope convex hull.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="show_scope_chull_frames",
                                         name="Scope convex hull frames: ",
                                         default_values=[0],
                                         help_text="Allows to indicate for which frames of object defined by \"Show scope convex hull\" should be displayed. It is possible to set several frames. In that case frames would be displayed as states.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="show_object_chull",
                                         name="Show object convex hull: ",
                                         default_values=[str()],
                                         help_text="If is set to selection of some molecular object in the system convex hull of this object is displayed. This works exacly the same way as show_chull but is meant to mark object shape. It can be achieved by using name * and molecular object definition plus some spatial constrains, for example those used in object definition.",
                                         level=2))
visualize_section.add_entry(DefaultEntry(config_name="show_object_chull_frames",
                                         name="Object convex hull frames:",
                                         default_values=[0],
                                         help_text="Allows to indicate for which frames of object defined by show_object convex hull should be displayed. It is possible to set several frames. In that case frames would be displayed as states.",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="cluster_area",
                                         name="Cluster area: ",
                                         default_values=[False],
                                         help_text="If set True, clusters’ areas are estimated with kernel density estimation method (KDE) and plotted as countour.",
                                         level=2))
visualize_section.add_entry(DefaultEntry(config_name="cluster_area_precision",
                                         name="Cluster area precision: ",
                                         default_values=[20],
                                         help_text="Precision of KDE method in clusters’ areas estimation method. This options controls number of grid points per one square A as used in KDE. Higher values means better precision. Number of points can be calculated as P^(2/3).",
                                         level=1))
visualize_section.add_entry(DefaultEntry(config_name="cluster_area_expand",
                                         name="Expand cluster area: ",
                                         default_values=[2],
                                         help_text="Space occupied by clusters’ points can be expanded before KDE calculation. This option controls amount of A by which the cluster space is expanded. Average amount of expansion can be calcualted as E^(2/3).",
                                         level=1))
DEFAULTS.append(visualize_section)

smooth_section = DefaultSection(config_name="smooth", name="Smooth", level=1)
smooth_section.add_entry(DefaultEntry(config_name="method",
                                      name="Method: ",
                                      default_values=[("window", "mss", "window_mss", "awin", "awin_mss", "dwin",
                                                       "dwin_mss", "savgol")],
                                      help_text="Smoothing method.",
                                      level=0))
smooth_section.add_entry(DefaultEntry(config_name="recursive",
                                      name="Recursive runs: ",
                                      default_values=[int()],
                                      help_text="Number of recursive runs of smoothing method.",
                                      level=0))
smooth_section.add_entry(DefaultEntry(config_name="window",
                                      name="Window size: ",
                                      default_values=[float()],
                                      help_text="In window based method defines window size. In plain window it has to be int number. In savgol it has to be odd integer.",
                                      level=0))
smooth_section.add_entry(DefaultEntry(config_name="step",
                                      name="Step: ",
                                      default_values=[float()],
                                      help_text="In step based method defines size of the step.",
                                      level=0))
smooth_section.add_entry(DefaultEntry(config_name="function",
                                      name="Averaging function: ",
                                      default_values=[("mean", "median")],
                                      help_text="In window based methods defines averaging function.",
                                      level=0))
smooth_section.add_entry(DefaultEntry(config_name="polyorder",
                                      name="Polyorder: ",
                                      default_values=[int()],
                                      help_text="In savgol is polynomial order.",
                                      level=0))

separate_paths_section.add_entry(smooth_section)

# Added rest of easy clustering section entries
clust_easy_section.add_entry(copy.copy(clustering_section))

clust_easy_section.add_entry(DefaultEntry(config_name="join_clusters",
                                          name="Join clusters: ",
                                          default_values=[str()],
                                          help_text="This option allows to join selected clusters. Clusters’ IDs joined with + character lists clusters to be joined together. Several such blocks separated by space can be used. For example, if set to 1+3+4 5+6 clusters 1, 3, and 4 will be joined in one cluster and cluster 5, and 6 will be also joined in another cluster.",
                                          level=2,
                                          group_label="Post Clustering Optimization"))
clust_easy_section.add_entry(DefaultEntry(config_name="renumber_clusters",
                                          name="Renumber clusters: ",
                                          default_values=[False],
                                          help_text="If set True, clusters have consecutive numbers starting from 1 (or 0 if outliers are present) starting from the bigest cluster.",
                                          level=2,
                                          group_label="Post Clustering Optimization"))

VALVE_DEFAULTS = DefaultSection("", "", 0)
VALVE_DEFAULTS.add_entry(DefaultEntry(config_name="-c",
                                      name="Config absolute file path: ",
                                      default_values=[filetype()],
                                      help_text="Absolute path to config file.",
                                      level=None
                                      ))
VALVE_DEFAULTS.add_entry(DefaultEntry(config_name="-t",
                                      name="Max threads: ",
                                      default_values=[str()],
                                      help_text="Limit Aqua-Duct calculations to given number of threads.",
                                      level=None
                                      ))
# VALVE_DEFAULTS.add_entry(DefaultEntry(config_name="--force-save",
#                                       name="Force saving results: ",
#                                       default_values=[False],
#                                       help_text="Force saving results.",
#                                       level=None
#                                       ))
VALVE_DEFAULTS.add_entry(DefaultEntry(config_name="--debug",
                                      name="Debug mode: ",
                                      default_values=[False],
                                      help_text="Prints debug info.",
                                      level=None
                                      ))
VALVE_DEFAULTS.add_entry(DefaultEntry(config_name="--debug-file",
                                      name="Debug absolute file path: ",
                                      default_values=[filetype()],
                                      help_text="Absolute path to debug file.",
                                      level=None
                                      ))

POND_DEFAULTS = DefaultSection("", "", 0)
POND_DEFAULTS.add_entry(DefaultEntry(config_name="-c",
                                     name="Config filename: ",
                                     default_values=[str()],
                                     help_text="Config file filename.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="-t",
                                     name="Max threads: ",
                                     default_values=[str()],
                                     help_text="Limit Aqua-Duct calculations to given number of threads.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="-r",
                                     name="Results directory: ",
                                     default_values=[str()],
                                     help_text="Path to results directory.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--paths-types",
                                     name="Paths types: ",
                                     default_values=[str()],
                                     help_text="Limit calculations to given paths types, i.e. given molecules.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--raw",
                                     name="Use raw paths: ",
                                     default_values=[str()],
                                     help_text="Use raw data from paths instead of single paths.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--raw-master",
                                     name="Use raw master paths: ",
                                     default_values=[False],
                                     help_text="Use raw data from paths instead of single paths, only in master paths calculations.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--raw-discard-singletons",
                                     name="Discard raw singletons: ",
                                     default_values=[str()],
                                     help_text="Discard short scope only segments from raw data.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--window-full",
                                     name="Full window: ",
                                     default_values=[False],
                                     help_text="Return full window if windows is used.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--windows",
                                     name="Windows: ",
                                     default_values=[str()],
                                     help_text="Number of windows to calculate.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--wsize",
                                     name="Window size: ",
                                     default_values=[str()],
                                     help_text="Size of window in frames.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--reference",
                                     name="Reference: ",
                                     default_values=[str()],
                                     help_text="Selection of reference in the first frame of trajectory.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--reference-radius",
                                     name="Reference radius: ",
                                     default_values=[str()],
                                     help_text="Radius of reference.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--reference-mol",
                                     name="Molecule reference: ",
                                     default_values=[str()],
                                     help_text="Selection of reference molecules.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--temperature",
                                     name="Temperature: ",
                                     default_values=[str()],
                                     help_text="Simulation temperature.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--gsize",
                                     name="Grid cells size: ",
                                     default_values=[str()],
                                     help_text="Size of grid's cells.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--pockets",
                                     name="Pockets: ",
                                     default_values=[False],
                                     help_text="Calculate pockets.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--hotspots",
                                     name="Hotspots: ",
                                     default_values=[False],
                                     help_text="Calculates hotspots if pockets are calculated.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--master-radius",
                                     name="Master paths radius: ",
                                     default_values=[False],
                                     help_text="Calculate profiles for master paths with given radius.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--master-ctypes",
                                     name="Ctypes: ",
                                     default_values=[False],
                                     help_text="Limit calculations to given ctypes.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--debug",
                                     name="Debug mode: ",
                                     default_values=[False],
                                     help_text="Prints debug info.",
                                     level=None
                                     ))
POND_DEFAULTS.add_entry(DefaultEntry(config_name="--debug-file",
                                     name="Debug mode: ",
                                     default_values=[False],
                                     help_text="Debug log file.",
                                     level=None
                                     ))
# @formatter:on

MENUS = [
    # section:entry
    "clustering:method",
    "reclustering:method"
]

LEVELS = {
    "Easy": 2,
    "Normal": 1,
    "Expert": 0
}



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Module performs HDR 2D calculations only with Gaussian Kerneld Density Estimator
as impelemented in :mod:`scipy.stats`.
'''

from aquaduct import logger

iscontour = True
try:
    from matplotlib import _contour
except ImportError:
    logger.warning("Cannot import _contour from matplotlib, contour calculations will be skipped.")
    iscontour = False


def hdr2contour(hdr, fraction=0.9):
    # X,Y,Z,no mask,corner mask,nchunk = 0
    _c = _contour.QuadContourGenerator(hdr.X, hdr.Y, hdr.Z, None, True, 0)
    cc = _c.create_contour(hdr.Z.max() * (1 - fraction))
    if len(cc) == 0:
        return
    return hdr.pca.undo(cc[0], pc=[0, 1])



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Aqua-Duct - a collection of tools to trace residues in MD simulation.
"""

import logging

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logger_name = __name__
logger = logging.getLogger(logger_name)
logger.addHandler(NullHandler())
logger.setLevel(logging.DEBUG)


def version():
    """
    Returns :mod:`aquaduct` version number.

    :return: 3 element tuple of int numbers
    :rtype: tuple
    """
    return 1, 0, 11


def version_nice():
    """
    Returns :mod:`aquaduct` version number as nicely formatted string.

    :return: string composed on the basis of the number returned by :func:`version`.
    :rtype: str
    """

    return '.'.join(map(str, version()))


__version__ = version_nice()
__mail__ = 'info@aquaduct.pl'

__author__ = 'Tomasz Magdziarz, Michał Banas, Alicja Płuciennik, Michał Stolarczyk'
__author_doc__ = 'Tomasz Magdziarz, Karolina Mitusińska, Agata Raczyńska, Artur Góra'


def greetings():
    """
    Returns fancy greetings of :mod:`aquaduct`. It has a form of ASCII-like
    graphic. Currently it returns following string::

        ------------------------------------------------
                  ~ ~ ~ A Q U A - D U C T ~ ~ ~
        ################################################
        ####        ########        ########        ####
        =@            \%%/            (==)            @=
        #              ##              ##              #
        #              ##              ##              #
        #              ##              ##              #
        #              ##              ##              #
        ------------------------------------------------


    :return: :mod:`aquaduct` fancy greetings.
    :rtype: str
    """
    greet = '''------------------------------------------------
          ~ ~ ~ A Q U A - D U C T ~ ~ ~
################################################
####        ########        ########        ####
=@            \%%/            (==)            @=
#              ##              ##              #
#              ##              ##              #
#              ##              ##              #
#              ##              ##              #
------------------------------------------------'''
    #          C l a s s i c a l  o r d e r
    # ------------------------------------------------'''
    return greet



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
What have I got in my pocket?
"""



################################################################################
# reuse AQ logger

from __future__ import absolute_import
from __future__ import print_function
import logging
from aquaduct import logger, logger_name

formatter_string = '%(name)s:%(levelname)s:[%(module)s|%(funcName)s@%(lineno)d]: %(message)s'
# create and add console handler with WARNING level to the AQ logger
formatter = logging.Formatter(formatter_string)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
ch.setLevel(logging.WARNING)  # default level is WARNING
logger.addHandler(ch)

################################################################################

import json
import gzip
import csv
import sys
from collections import namedtuple

################################################################################

# Boltzmann k constant
# k = 0.0019872041 # kcal/mol/K
k = 0.0083144621  # kJ/mol/K
k_unit = 'kJ/mol/K'

################################################################################

class count_ref(object):

    def __init__(self, pbar, ref_sel):
        self.pbar = pbar  # queue
        self.ref_sel = ref_sel

    def __call__(self, traj_reader):
        traj_reader = traj_reader.open()
        ref = 0.
        for frame in traj_reader.iterate():
            ref += len(list(traj_reader.parse_selection(self.ref_sel).residues().ids()))
            self.pbar.put(1)  # report one frame a time
        return ref


################################################################################


if __name__ == "__main__":

    from sys import exc_info

    try:
        from aquaduct.utils import clui

        with clui.tictoc('What have I got in my pocket?'):

            # ----------------------------------------------------------------------#
            # argument parsing

            import argparse
            from aquaduct import version_nice as aquaduct_version_nice

            description_version = '''Aquaduct library version %s''' % (aquaduct_version_nice(),)
            description = '''What have I got in my pocket?'''

            parser = argparse.ArgumentParser(description=description,
                                             formatter_class=argparse.ArgumentDefaultsHelpFormatter)

            parser.add_argument("-c", action="store", dest="config_file", required=False,
                                help="Config file filename.")
            parser.add_argument("-t", action="store", dest="threads", required=False, default=None,
                                help="Limit Aqua-Duct calculations to given number of threads.")
            parser.add_argument("-r", action="store", dest="results_dir", required=False,
                                help="Path to results directory", default="", type=str)
            parser.add_argument("--debug", action="store_true", dest="debug", required=False,
                                help="Prints debug info.")
            parser.add_argument("--debug-file", action="store", dest="debug_file", required=False,
                                help="Debug log file.")
            parser.add_argument("--paths-types", action="store", dest="paths_types", type=str, required=False,
                                default="",
                                help="Limit calculations to given paths types, i.e. given molecules.")
            parser.add_argument("--raw", action="store_true", dest="raw", required=False,
                                help="Use raw data from paths instead of single paths.")
            parser.add_argument("--raw-master", action="store_true", dest="raw_master", required=False,
                                help="Use raw data from paths instead of single paths, only in master paths calculations.")
            parser.add_argument("--raw-discard-singletons", action="store", dest="raw_singl", required=False, type=int,
                                default=1,
                                help="Discard short scope only segments from raw data.")
            parser.add_argument("--window-full", action="store_true", dest="wfull", required=False,
                                help="Return full window if windows is used.")
            parser.add_argument("--windows", action="store", dest="windows", type=int, required=False, default=1,
                                help="Number of windows to calculate.")
            parser.add_argument("--wsize", action="store", dest="wsize", type=int, required=False, default=None,
                                help="Size of window in frames.")
            parser.add_argument("--reference-value", action="store", dest="ref", type=float, required=False, default=None,
                                help="Reference value in [%s]." % k_unit)
            parser.add_argument("--reference-calc", action="store_true", dest="ref_calc", required=False,
                                help="Calculate reference value with scope and reference molecules.")
            parser.add_argument("--reference-radius", action="store", dest="ref_radius", type=float, required=False,
                                default=2.,
                                help="Radius of reference in [Å].")
            parser.add_argument("--reference-mol", action="store", dest="ref_mol", type=str, required=False,
                                default='resname WAT',
                                help="Selection of reference molecules.")
            parser.add_argument("--temperature", action="store", dest="temp", type=float, required=False, default=300.,
                                help="Simulation temperature in [K].")
            parser.add_argument("--gsize", action="store", dest="grid_size", type=float, required=False, default=1.,
                                help="Size of grid's cells in in [Å].")
            parser.add_argument("--pockets", action="store_true", dest="pockets", required=False,
                                help="Calculate pockets.")
            parser.add_argument("--hotspots", action="store_true", dest="hotspots", required=False,
                                help="Calculates hotspots if pockets are calculated.")
            parser.add_argument("--energy-profile", action="store_true", dest="eng_pro", required=False,
                                help="Calculates energy profiles for master paths.")
            # FIXME: Temporary option for loading master paths when they are not needed due to master_radius default value.
            parser.add_argument("--master", action="store_true", dest="master", required=False,
                                help="Enables master paths calculation.")
            parser.add_argument("--master-radius", action="store", dest="master_radius", type=float, required=False, default=2.,
                                help="Calculate profiles for master paths with given radius in [Å].")
            parser.add_argument("--master-ctypes", action="store", dest="master_ctypes", type=str, required=False,
                                default="",
                                help="Limit calculations to given ctypes.")
            # parser.add_argument("--master-radii", action="store_true", dest="master_radii", required=False,
            #                    help="Calculate profiles for master paths using width as radii.")
            parser.add_argument("--io-threshold", action="store", dest="io_threshold", type=float, required=False,
                                default=None,
                                help="Percent value of maximal density which will be used to partition pocket into inner and outer instead of mean value.")
            parser.add_argument("--path-id", action="store", dest="path_id", type=str, required=False,
                                default=None, help="Calculate profiles for specified path ID.")
            parser.add_argument("--path-file", action="store", dest="path_file", type=str, required=False,
                                help="Use coordinates from specified CSV file.")
            parser.add_argument("--path-radius", action="store", dest="path_radius", type=float, required=False,
                                default=2., help="Calculate profiles for path with given radius in [Å].")
            parser.add_argument("--path-smooth", action="store_true", dest="path_smooth", required=False,
                                help="If used path coordinates will be smoothed.")
            parser.add_argument("--raw-path", action="store_true", dest="raw_path", required=False,
                                help="Use raw data from paths instead of single paths. "
                                     "Used for path energy profiles calculations and for extracting raw path.")
            parser.add_argument("--extract-path", action="store", dest="extract_path", type=str, required=False,
                                default=None, help="Extract path coordinates with specified ID.")
            parser.add_argument("--output-file", action="store", dest="output_file", type=str, required=False,
                                default="path_coords.csv", help="Output CSV filename for extracted coordinates.")
            parser.add_argument("--output-suffix", action="store", dest="output_suffix", type=str, required=False, default="", help="Arbitrary suffix appended to output filenames.")


            args = parser.parse_args()

            # ----------------------------------------------------------------------#

            # debug
            # at this stage logger is the AQ root logger
            if args.debug:
                logger.removeHandler(ch)  # remove old ch handlers
                CH = logging.StreamHandler()
                ch.setFormatter(formatter)
                ch.setLevel(logging.DEBUG)
                logger.addHandler(ch)
            if args.debug_file:
                formatter = logging.Formatter('%(asctime)s: ' + formatter_string)
                fh = logging.FileHandler(args.debug_file)
                fh.setFormatter(formatter)
                fh.setLevel(logging.DEBUG)
                logger.addHandler(fh)
            # finally, get valve logger
            logger = logging.getLogger(logger_name + '.pond')
            logger.info('Initialization of Pond logging done.')

            # ----------------------------------------------------------------------#
            # config

            from aquaduct.apps.valve import ValveConfig, valve_load_config

            config = ValveConfig()  # config template
            valve_load_config(args.config_file, config)
            # get global options
            goptions = config.get_global_options()

            # ----------------------------------------------------------------------#
            # init and options

            from aquaduct.utils.multip import optimal_threads
            from aquaduct.traj.sandwich import Reader, Window
            from aquaduct.apps.valve.data import get_vda_reader
            from aquaduct.apps.data import save_cric
            from os import pathsep
            import numpy as np
            from multiprocessing import Pool, Manager
            from aquaduct.geom import pocket
            from aquaduct.traj.dumps import WriteMOL2
            from aquaduct.apps.valve.helpers import get_linearize_method
            from aquaduct.apps.valve.helpers import get_smooth_method
            from aquaduct.geom import traces
            
            import os
            from aquaduct.geom.smooth import SavgolSmooth

            if args.threads is None:
                optimal_threads.threads_count = optimal_threads.cpu_count + 1
            else:
                optimal_threads.threads_count = int(args.threads)
            clui.message("Number of threads Pond is allowed to use: %d" % optimal_threads.threads_count)
            if (1 < optimal_threads.threads_count < 3) or (
                    optimal_threads.threads_count - 1 > optimal_threads.cpu_count):
                clui.message(
                    "Number of threads is not optimal; CPU count reported by system: %d" % optimal_threads.cpu_count)
            # because it is used by mp.Pool it should be -1???
            if optimal_threads.threads_count > 1:
                optimal_threads.threads_count -= 1
                clui.message("Main process would use 1 thread.")
                clui.message("Concurent calculations would use %d threads." % optimal_threads.threads_count)

            # Maximal frame checks
            frames_window = Window(goptions.min_frame, goptions.max_frame, goptions.step_frame)

            # Open trajectory reader
            # trajectories are split with os.pathsep

            Reader(goptions.top, goptions.trj,
                   window=frames_window,
                   sandwich=goptions.sandwich,
                   threads=optimal_threads.threads_count)  # trajectory reader

            # ----------------------------------------------------------------------#
            # results dir
            rdir = ""
            if len(args.results_dir):
                if not os.path.exists(args.results_dir):
                    os.makedirs(args.results_dir)
                rdir = args.results_dir
            if len(rdir) == 0:
                rdir += '.'
            rdir += os.path.sep

            results_meta = {'options': vars(args)}
            rmu = lambda k, v: results_meta.update({k: v})
            if args.output_suffix:
                args.output_suffix = '_%s' % args.output_suffix
            else: 
                args.output_suffix = ''

            # ----------------------------------------------------------------------#
            # load paths
            paths_types = [pt.strip() for pt in args.paths_types.split(' ') if len(pt.strip())]
            ptn = ''  # additiona paths types name suffix
            if paths_types:
                rmu('paths_types', paths_types)
                clui.message('Limiting calculations to paths of %s.' % ', '.join(paths_types))
                ptn = '_' + '_'.join(paths_types)

            if (args.pockets or args.master_radius) and not (args.master_radius and not args.raw and args.raw_master):
                with clui.tictoc('Loading paths'):
                    if not args.raw:
                        # get stage III options
                        options3 = config.get_stage_options(2)
                        with clui.fbm('Loading data dump from %s file' % options3.dump):
                            vda = get_vda_reader(options3.dump, mode='r')
                            result3 = vda.load()
                        paths = result3.pop('spaths')
                        if paths_types:
                            paths = [p for p in paths if p.id.name in paths_types]
                        rmu('paths', 'spaths')
                    else:
                        # get stage II options
                        options2 = config.get_stage_options(1)
                        with clui.fbm('Loading data dump from %s file' % options2.dump):
                            vda = get_vda_reader(options2.dump, mode='r')
                            result2 = vda.load()
                        with clui.fbm('Preparing raw data'):
                            paths = result2.pop('paths')
                            if args.raw_singl:
                                for p in paths:
                                    p.discard_singletons(singl=args.raw_singl)
                            paths = [p for p in paths if len(p.frames)]
                            if paths_types:
                                paths = [p for p in paths if p.name in paths_types]
                        rmu('paths', 'paths')

            # ----------------------------------------------------------------------#
            # reference value

            ref = args.ref
            # calculate n_z for reference
            if args.ref_calc:
                with clui.tictoc('Reference calculation'):
                    from aquaduct.geom import traces
                    from scipy.spatial.distance import cdist
                    # 1. Get scope atoms.
                    # 2. Calculate convexhull.
                    # 3. Get first vertex and in the direction in which it is oriented look for farthest reference molecule.
                    # 4. Point for reference calculation get as midpoint between the used vertex and the farthest molecule.
                    # 5. Proceed normally.
                    options2 = config.get_stage_options(1)

                    for traj_reader in Reader.iterate():
                        traj_reader = traj_reader.open()
                        for frame in traj_reader.iterate():
                            with clui.pbar(mess="Automatic reference point calculation",maxval=7) as pbar:
                                # 1.
                                scope = traj_reader.parse_selection(options2.scope)
                                pbar.next()
                                # 2.
                                scope_ch = scope.chull(inflate=options2.scope_convexhull_inflate)
                                pbar.next()
                                # 3.
                                v = scope_ch.vertices_points[0] # vertex
                                c = np.mean(scope_ch.vertices_points) # center of scope
                                #ref_mol = traj_reader.parse_selection(args.ref_mol).residues().coords()
                                ref_mol = traj_reader.select_all().residues().coords()
                                ref_mol = np.array(list(ref_mol)) # reference molecules coordinates
                                pbar.next()
                                # find all ref_mol that are above the vertex
                                ref_mol = ref_mol[[traces.is_p_above_vp0_plane(rm,v-c,v) > 0 for rm in ref_mol]]
                                pbar.next()
                                # find distances of remaining ref_mol to vc line
                                rmd = [traces.distance_p_to_ab(rm,v,c) for rm in ref_mol]
                                pbar.next()
                                # find all ref_mol that are within reference radius from vc line
                                ref_mol = ref_mol[[d < args.ref_radius for d in rmd]]
                                pbar.next()
                                # find the most distant ref_mol
                                com = cdist(ref_mol,[v])
                                com = ref_mol[np.argmax(com)]
                                pbar.next()
                            # 4.
                            clui.message("Using reference point as middle between (%0.4f,%0.4f,%0.4f) and (%0.4f,%0.4f,%0.4f)." % (tuple(map(float,v))+tuple(map(float,com))))
                            com = (com+v)/2
                            clui.message("Using reference point (%0.4f,%0.4f,%0.4f)." % tuple(map(float,com)))
                            break
                        break
                    Reader.reset()

                    goal = Reader.number_of_frames(onelayer=False)
                    with clui.pbar(goal, mess="Calculating density in the reference area:") as pbar:
                        ref = []
                        ref_sel = '(%s) and (point %f %f %f %f)' % (
                                (args.ref_mol,) + tuple(map(float, com)) + (args.ref_radius,))
                        # ref_sel = '(%s) and (sphzone %f %s)' % (args.ref_mol, args.ref_radius, args.ref)
                        manager = Manager()
                        pbar_queue = manager.Queue()
                        pool = Pool(processes=optimal_threads.threads_count)

                        r = pool.map_async(
                            count_ref(pbar_queue, ref_sel),
                            Reader.iterate(),
                            callback=ref.extend)

                        progress = 0
                        for p in iter(pbar_queue.get, None):
                            progress += p
                            pbar.next(step=p)
                            if progress == goal:
                                break
                        r.wait()
                        pool.close()
                        pool.join()
                        del pbar_queue, manager

                    ref = float(sum(ref))

                    clui.message('Reference number of molecules: %d [molecules].' % int(ref))

                    ref /= float(Reader.number_of_frames(onelayer=False))  # if sandwich mean value will be correct

                    ref /= 4. / 3. * np.pi * (float(args.ref_radius) ** 3)
                    clui.message('Reference density: %0.4f [molecules/A^3].' % ref)

                    # Boltzmann k constant
                    # k = 0.0019872041 # kcal/mol/K
                    # k = 0.0083144621  # kJ/mol/K
                    rmu('energy_unit', k_unit)

                    ref = -k * args.temp * np.log(ref)
                    rmu('reference_correction', float(ref))
                    clui.message('Reference value: %0.4f [kJ/mol/K].' % ref)

            if ref:
                rmu('reference_density_correction', np.exp(ref/(-k * args.temp)))

            # ----------------------------------------------------------------------#
            # windows

            W = args.windows
            WS = args.wsize

            many_windows = W > 1 or (W == 1 and (WS is not None))
            many_windows = many_windows and WS < Reader.number_of_frames(onelayer=True)

            windows = []
            clui.message("Calculation will be done for following windows:")
            for wnr, window in enumerate(pocket.windows(Reader.number_of_frames(onelayer=True), windows=W, size=WS)):
                middle = sum(window) / 2
                if wnr and many_windows:
                    print(("W%d %d:%d middle: %d" % (wnr, window[0], window[1], middle)))
                    windows.append([wnr, window[0], window[1], middle])
                elif (wnr == 0) and (args.wfull or (not many_windows)):
                    print(("full %d:%d" % (window[0], window[1])))
                    rmu('full_window',[window[0], window[1]])
                rmu('windows', windows)

            # ----------------------------------------------------------------------#
            # calculate pockets

            if args.pockets:
                clui.message('What have I got in my pocket?')
                with clui.tictoc('Pockets & hotspots calculation'):

                    W = args.windows
                    WS = args.wsize
                    grid_size = args.grid_size
                    grid_area = grid_size ** 3
                    with clui.pbar(len(paths) * (1 + W + int(args.wfull)), mess='Calculating pockets:') as pbar:
                        pockets_volume = open((rdir + ('volumes%s%s.dat' % (ptn, args.output_suffix))), 'w')
                        pockets_volume.write(('\t'.join('W_start W_end Outer Inner'.split())) + os.linesep)
                        pool = Pool(processes=optimal_threads.threads_count)
                        edges = pocket.find_edges(paths, grid_size=grid_size, pbar=pbar, map_fun=pool.imap_unordered)
                        number_of_frames = Reader.number_of_frames(onelayer=True)
                        if WS is None:
                            WSf = float(number_of_frames / float(W))
                        else:
                            WSf = float(WS)
                        if Reader.sandwich_mode:
                            WSf *= Reader.number_of_layers()
                        wmol2 = None
                        hsmol2 = None
                        for wnr, window in enumerate(
                                pocket.windows(Reader.number_of_frames(onelayer=True), windows=W, size=WS)):
                            number_of_frames = (window[-1] - window[0])
                            if Reader.sandwich_mode:
                                number_of_frames *= Reader.number_of_layers()

                            if wnr and many_windows:
                                if wmol2 is None:
                                    wmol2 = [WriteMOL2(rdir + 'outer%s%s.mol2' % (ptn, args.output_suffix)),
                                             WriteMOL2(rdir + 'inner%s%s.mol2' % (ptn, args.output_suffix))]
                                if hsmol2 is None and args.hotspots:
                                    hsmol2 = WriteMOL2(rdir + 'hotspots%s%s.mol2' % (ptn, args.output_suffix))
                                D = pocket.distribution(paths, grid_size=grid_size, edges=edges, window=window,
                                                        pbar=pbar, map_fun=pool.imap_unordered)
                                H = (D[-1] / WSf) / grid_area

                                if args.hotspots:
                                    hs = pocket.hot_spots(H)
                                    if hs is not None:
                                        hs = H >= hs
                                        hsmol2.write_scatter(D[0][hs], H[hs])
                                    else:
                                        hsmol2.write_scatter([], [])

                                volumes = []
                                for I, mol2 in zip(pocket.outer_inner(D[-1], args.io_threshold), wmol2):
                                    mol2.write_scatter(D[0][I], H[I])
                                    volumes.append(sum(I) * grid_area)
                                pockets_volume.write(('%d\t%d\t%0.1f\t%0.1f' % (window + tuple(volumes))) + os.linesep)
                            #elif args.wfull:
                            elif (wnr == 0) and (args.wfull or (not many_windows)):
                                D = pocket.distribution(paths, grid_size=grid_size, edges=edges, window=window,
                                                        pbar=pbar, map_fun=pool.imap_unordered)
                                H = (D[-1] / float(number_of_frames)) / grid_area

                                if args.hotspots:
                                    hs = pocket.hot_spots(H)
                                    mol2 = WriteMOL2(rdir + 'hotspots_full%s%s.mol2' % (ptn, args.output_suffix))
                                    if hs is not None:
                                        hs = H >= hs
                                        mol2.write_scatter(D[0][hs], H[hs])
                                    else:
                                        mol2.write_scatter([], [])
                                    del mol2

                                volumes = []
                                for I, mol2 in zip(pocket.outer_inner(D[-1], args.io_threshold),
                                                   [WriteMOL2(rdir + 'outer_full%s%s.mol2' % (ptn, args.output_suffix)),
                                                    WriteMOL2(rdir + 'inner_full%s%s.mol2' % (ptn, args.output_suffix))]):
                                    mol2.write_scatter(D[0][I], H[I])
                                    volumes.append(sum(I) * grid_area)
                                    del mol2
                                pockets_volume.write(('%d\t%d\t%0.1f\t%0.1f' % (window + tuple(volumes))) + os.linesep)
                            save_cric()
                        pool.close()
                        pool.join()

                        if wmol2 is not None:
                            for mol2 in wmol2:
                                del mol2
                        if hsmol2 is not None:
                            if args.hotspots:
                                del hsmol2
                        pockets_volume.close()

                clui.message("what it's got in its nassty little pocketses?")

            # ----------------------------------------------------------------------#
            # load mater paths data
            if args.master:
                if args.master_radius and args.master:
                    with clui.tictoc('Loading master paths data'):
                        # get stage IV options
                        options4 = config.get_stage_options(3)

                        with clui.fbm('Loading data dump from %s file' % options4.dump):
                            vda = get_vda_reader(options4.dump, mode='r')
                            result4 = vda.load()
                            mps = result4.pop(
                                'master_paths_smooth')  # FIXME: let user decide what kind of master paths are used

                        options6 = config.get_stage_options(5)

                        linmet = get_linearize_method(options6.simply_smooths)
                else:
                    # FIXME: Temporary solution for loading master paths when they are not needed due to master_radius default value.
                    mps = {}

                # ----------------------------------------------------------------------#
                # re load paths?

                if args.master_radius and args.raw_master and (not args.raw):
                    with clui.tictoc('ReLoading paths (raw data)'):
                        # get stage II options
                        options2 = config.get_stage_options(1)
                        with clui.fbm('Loading data dump from %s file' % options2.dump):
                            vda = get_vda_reader(options2.dump, mode='r')
                            result2 = vda.load()
                        with clui.fbm('Preparing raw data'):
                            paths = result2.pop('paths')
                            if args.raw_singl:
                                for p in paths:
                                    p.discard_singletons(singl=args.raw_singl)
                            paths = [p for p in paths if len(p.frames)]
                            if paths_types:
                                paths = [p for p in paths if p.name in paths_types]

                # ----------------------------------------------------------------------#
                # master paths profiles

                limit_ctypes = [ct.strip() for ct in args.master_ctypes.split(' ')]
                if limit_ctypes != [""]:
                    clui.message('Limiting master paths data to %s ctypes.' % (' '.join(limit_ctypes)))
                    for ctk in list(mps.keys()):
                        if str(ctk) not in limit_ctypes:
                            mps.pop(ctk)

                if args.master_radius and len(mps) == 0:
                    clui.message("No master paths data.")

                if args.eng_pro and len(mps) and ref is not None:
                    with clui.tictoc('Master paths profiles calculation'):

                        limit_ctypes = [ct.strip() for ct in args.master_ctypes.split(' ')]
                        if limit_ctypes != [""]:
                            for ctk in list(mps.keys()):
                                if str(ctk) not in limit_ctypes:
                                    mps.pop(ctk)

                        W = args.windows
                        WS = args.wsize

                        many_windows = W > 1 or (W == 1 and (WS is not None))
                        many_windows = many_windows and WS < Reader.number_of_frames(onelayer=True)

                        for wnr, window in enumerate(
                                pocket.windows(Reader.number_of_frames(onelayer=True), windows=W, size=WS)):

                            number_of_frames = (window[-1] - window[0])
                            if Reader.sandwich_mode:
                                number_of_frames *= Reader.number_of_layers()

                            if wnr or args.wfull:  # or not many_windows:
                                for ctype, mp in mps.items():
                                    if isinstance(mp,dict):
                                        logger.warning("Pond cannot yet handle MasterPaths calculated with separate_master option.")
                                        logger.warning("MasterPaths for %s skip." % ctype)
                                        continue

                                    fname_window = ""
                                    fname_window_single = ""
                                    fname = str(ctype).replace(':', '-')
                                    if not wnr and args.wfull:
                                        fname += 'full'
                                    elif wnr and many_windows:
                                        fname_window = '_W%d' % wnr
                                        fname_window_single = "_WX"

                                    window_name = fname if fname else wnr
                                    with clui.pbar(mess="Calculate energy for window {}".format(window_name),
                                                   maxval=int(number_of_frames)) as pbar:

                                        pool = Pool(processes=optimal_threads.threads_count)

                                        mode = 'w'
                                        if wnr > 1:
                                            mode = 'a'

                                        centers, ids = linmet(mp.coords_cont, ids=True)

                                        D = pocket.sphere_density_raw(Reader.iterate(), args.ref_mol,
                                                                      centers, args.master_radius, pool,
                                                                      window=window, pbar=pbar)

                                        H = D / float(number_of_frames) / (4. / 3. * np.pi * float(args.master_radius) ** 3)
                                        if ref:
                                            H = -k * args.temp * np.log(H) - ref

                                        with WriteMOL2(rdir + "mp_%s%s_radius%s%s.mol2" % (fname, fname_window_single, ptn, args.output_suffix),
                                                       mode=mode) as mol2:
                                            mol2.write_connected(centers, H)

                                        with open(rdir + "mp_%s%s_radius.dat" % (fname, fname_window), 'w') as dat:
                                            dat.write('len\tE' + os.linesep)
                                            L = np.hstack((0., np.cumsum(traces.diff(centers))))
                                            for l, E in zip(L, H):
                                                dat.write('%f\t%f%s' % (l, E, os.linesep))
                                    pool.close()
                                    pool.join()
                                save_cric()

            # ----------------------------------------------------------------------#
            # Paths
            # ----------------------------------------------------------------------#
            # Extracting

            def path_id_formatter(id):
                """
                Change tuple values to string with values separated by colon. Otherwise return string representation.
                :param id: tuple
                :return: string
                """
                if isinstance(id, tuple):
                    return ":".join([str(x) for x in id])
                else:
                    return str(id)

            if args.extract_path:
                if not args.raw_path:
                    options3 = config.get_stage_options(2)
                    with clui.fbm("Loading data dump from {} file.".format(options3.dump)):
                        vda = get_vda_reader(options3.dump, mode="r")
                        result3 = vda.load()

                        paths = result3.pop("spaths")
                else:
                    options2 = config.get_stage_options(1)
                    with clui.fbm('Loading raw data dump from %s file' % options2.dump):
                        vda = get_vda_reader(options2.dump, mode='r')
                        result2 = vda.load()

                        paths = result2.pop("paths")

                with clui.fbm("Extracting path {} coordinates.".format(args.extract_path)):
                    try:
                        path = next(path for path in paths if path_id_formatter(path.id) == str(args.extract_path))

                    except StopIteration:
                        path = None

                    if path:
                        if not args.raw_path:
                            soptions = result3.pop("soptions")
                            soptions = namedtuple('Options', list(soptions.keys()))(*list(soptions.values()))
                            smooth_method = get_smooth_method(soptions)

                            coords = path.get_coords_cont(smooth_method)
                        else:
                            coords = path.coords

                        with open(rdir + args.output_file, "w") as csvfile:
                            csv_writer = csv.writer(csvfile)

                            for coord in coords:
                                csv_writer.writerow(coord)
                    else:
                        clui.message("Path with ID {} does not exists.".format(args.extract_path))

                if path:
                    clui.message("Path saved to {}".format(args.output_file))

            # ----------------------------------------------------------------------#
            # Calculating energy profile for path

            if args.eng_pro and (args.path_id or args.path_file):
                with clui.tictoc('Loading paths data'):

                    if not args.raw_path:
                        options3 = config.get_stage_options(2)
                        with clui.fbm("Loading data dump from {} file.".format(options3.dump)):
                            vda = get_vda_reader(options3.dump, mode="r")
                            result3 = vda.load()

                            paths = result3.pop("spaths")
                    else:
                        options2 = config.get_stage_options(1)
                        with clui.fbm('Loading raw data dump from %s file' % options2.dump):
                            vda = get_vda_reader(options2.dump, mode='r')
                            result2 = vda.load()

                            paths = result2.pop("paths")

                    if args.path_id:
                        try:
                            path = next(path for path in paths if path_id_formatter(path.id) == args.path_id)

                            if not args.raw_path:
                                if args.path_smooth:
                                    soptions = result3.pop("soptions")
                                    soptions = namedtuple('Options', list(soptions.keys()))(*list(soptions.values()))
                                    smooth_method = get_smooth_method(soptions)
                                    coords = path.get_coords_cont(smooth_method)
                                else:
                                    coords = path.get_coords_cont()
                            else:
                                coords = path.coords
                        except StopIteration:
                            coords = []
                    elif args.path_file:
                        coords = []
                        with open(args.path_file, "r") as csvfile:
                            csv_reader = csv.reader(csvfile)

                            for row in csv_reader:
                                coords.append([float(r) for r in row])

                        coords = np.array(coords, dtype=np.float32)

                    if args.raw_path and args.raw_singl:
                        with clui.fbm('Preparing raw data'):
                            for p in paths:
                                p.discard_singletons(singl=args.raw_singl)
                        paths = [p for p in paths if len(p.frames)]

                    options6 = config.get_stage_options(5)
                    linmet = get_linearize_method(options6.simply_smooths)

                    coords = linmet(coords)

                if len(coords):
                    with clui.tictoc("Path profiles calculation", stdout=sys.stdout):

                        W = args.windows
                        WS = args.wsize

                        many_windows = W > 1 or (W == 1 and (WS is not None))
                        many_windows = many_windows and WS < Reader.number_of_frames(onelayer=True)

                        path_name = args.path_id if args.path_id else ""

                        for wnr, window in enumerate(
                                pocket.windows(Reader.number_of_frames(onelayer=True), windows=W, size=WS)):

                            number_of_frames = (window[-1] - window[0])
                            if Reader.sandwich_mode:
                                number_of_frames *= Reader.number_of_layers()

                            if wnr or args.wfull:  # or not many_windows:
                                # TODO: Handling names variables need to be refactored
                                fname_window = ""
                                fname_window_single = ""
                                fname = ""
                                if not wnr and args.wfull:
                                    window_name = "full"
                                    fname += 'full'
                                elif wnr and many_windows:
                                    fname_window = '_W%d' % wnr
                                    fname_window_single = "_WX"

                                window_name = fname if fname else wnr
                                with clui.pbar(mess="Calculate energy for window {}".format(window_name),
                                               maxval=int(number_of_frames)) as pbar:

                                    pool = Pool(processes=optimal_threads.threads_count)

                                    mode = 'w'
                                    if wnr > 1:
                                        mode = 'a'

                                    D = pocket.sphere_density_raw(Reader.iterate(), args.ref_mol,
                                                                  coords, args.path_radius, pool,
                                                                  window=window, pbar=pbar)

                                    H = D / float(number_of_frames) / (
                                            4. / 3. * np.pi * float(args.path_radius) ** 3)
                                    if ref:
                                        H = -k * args.temp * np.log(H) - ref

                                    with WriteMOL2(
                                            rdir + "path%s_%s%s_radius%s%s.mol2" % (
                                                    path_name, fname, fname_window_single, ptn, args.output_suffix),
                                            mode=mode) as mol2:
                                        mol2.write_connected(coords, H)

                                    with open(rdir + "path%s_%s%s_radius.dat" % (path_name, fname, fname_window),
                                              'w') as dat:
                                        dat.write('len\tE' + os.linesep)
                                        L = np.hstack((0., np.cumsum(traces.diff(coords))))
                                        for l, E in zip(L, H):
                                            dat.write('%f\t%f%s' % (l, E, os.linesep))

                                    pool.close()
                                    pool.join()
                            save_cric()
                else:
                    clui.message("Path with ID {} does not exists.".format(args.path_id))

            Reader.reset()
            with gzip.open(rdir + ('pond_meta%s%s.json' % (ptn, args.output_suffix)), mode='wt', compresslevel=9) as f:
                json.dump(results_meta, f)
                # TODO: consider usage of IterEncoder - move it to aquaduct/apps/data.py module

            # ----------------------------------------------------------------------#

    except BaseException:
        clui.emit_tvtb_to_file_in_root_logger(exc_info())
        raise



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from aquaduct import logger
import os


class WriteMOL2(object):

    def __init__(self, mol2file, mode='w'):

        self.current_atom = 1
        self.fh = open(mol2file, mode)

    def print_atom_line(self, xyz, bf):
        atom = "%7d  H         %3.4f   %3.4f    %3.4f H       1  FIL1        %3.4f" % (
            self.current_atom, xyz[0], xyz[1], xyz[2], bf)
        return atom + os.linesep

    def print_bond_line(self, bid, ba, bb, btype='1'):
        bond = "%6d%6d%6d%6s" % (bid, ba, bb, btype)
        return bond + os.linesep

    def write_scatter(self, scatter, bf):
        self.fh.write("@<TRIPOS>MOLECULE" + os.linesep)
        self.fh.write("pocket" + os.linesep)
        self.fh.write((" %d 0 0 0" % len(scatter)) + os.linesep)
        self.fh.write("SMALL" + os.linesep)
        self.fh.write("GASTEIGER" + os.linesep + os.linesep)
        self.fh.write("@<TRIPOS>ATOM" + os.linesep)
        for xyz, b in zip(scatter, bf):
            self.fh.write(self.print_atom_line(xyz, b))
            self.current_atom += 1
        self.fh.write("@<TRIPOS>BOND" + os.linesep)
        self.current_atom = 1

    def write_connected(self, scatter, bf):
        self.fh.write("@<TRIPOS>MOLECULE" + os.linesep)
        self.fh.write("pocket" + os.linesep)
        self.fh.write((" %d %d 0 0" % (len(scatter), len(scatter) - 1)) + os.linesep)
        self.fh.write("SMALL" + os.linesep)
        self.fh.write("GASTEIGER" + os.linesep + os.linesep)
        self.fh.write("@<TRIPOS>ATOM" + os.linesep)
        for xyz, b in zip(scatter, bf):
            self.fh.write(self.print_atom_line(xyz, b))
            self.current_atom += 1
        self.fh.write("@<TRIPOS>BOND" + os.linesep)
        for b in range(1, len(scatter)):
            self.fh.write(self.print_bond_line(b, b, b + 1))
        self.current_atom = 1

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        if typ is None:
            self.__del__()

    def __del__(self):
        self.fh.close()


class WritePDB(object):

    def __init__(self, pdbfile, csvfile=None, scale_bf=1):

        self.current_atom = 1
        self.scale_bf = scale_bf
        self.current_model = 0

        self.fh = open(pdbfile, 'w')
        self.fh_csv = None
        if csvfile:
            self.fh_csv = open(csvfile, 'w')

    def print_atom_line(self, xyz, bf):
        atom = "%5d" % self.current_atom
        x = ((" " * 8) + ("%0.3f" % xyz[0])[:8])[-8:]
        y = ((" " * 8) + ("%0.3f" % xyz[1])[:8])[-8:]
        z = ((" " * 8) + ("%0.3f" % xyz[2])[:8])[-8:]
        # x = ("        %0.3f" % xyz[0])[-8:]
        # y = ("        %0.3f" % xyz[1])[-8:]
        # z = ("        %0.3f" % xyz[2])[-8:]
        b = ((" " * 4) + ("%0.2f" % (bf * self.scale_bf))[:4])[-4:]
        return ("ATOM  %s  H   FIL T   1    %s%s%s        %s" % (atom, x, y, z, b)) + os.linesep

    def print_conect_line(self, a1, a2):
        atom1 = "%5d" % a1
        atom2 = "%5d" % a2
        return ("CONECT%s%s" % (atom1, atom2)) + os.linesep

    def write_connected(self, line, bf):
        conect = None
        for xyz, b in zip(line, bf):
            self.fh.write(self.print_atom_line(xyz, b))
            if conect:
                self.fh.write(self.print_conect_line(conect, self.current_atom))
            elif self.fh_csv is not None:
                self.fh_csv.write("X,Y,Z" + os.linesep)  # header
            if self.fh_csv is not None:
                self.fh_csv.write(("%f,%f,%f" + os.linesep) % tuple(xyz))
            conect = self.current_atom
            self.current_atom += 1
        if self.current_model:
            self.fh.write('ENDMDL' + os.linesep)

    def next_model(self):
        self.current_model += 1
        self.fh.write(('MODEL     %4d' % (self.current_model)) + os.linesep)
        self.current_atom = 1

    def write_scatter(self, scatter, bf):
        for xyz, b in zip(scatter, bf):
            self.fh.write(self.print_atom_line(xyz, b))
            self.current_atom += 1
        if self.current_model:
            self.fh.write('ENDMDL' + os.linesep)

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        if typ is None:
            self.__del__()

    def __del__(self):
        self.fh.close()
        if self.fh_csv is not None:
            self.fh_csv.close()



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pickle as pickle
import os
import tarfile
import json


import aquaduct.visual.pymol_cgo as cgo

from aquaduct.geom import traces
from aquaduct.traj.paths import PathTypesCodes
from aquaduct.utils.helpers import list_blocks_to_slices
from aquaduct.visual.helpers import color_codes, cc
from aquaduct.utils.helpers import create_tmpfile, listify


class BasicPymolCGO(object):
    cgo_entity_begin = []
    cgo_entity_end = []

    def __init__(self):
        self.cgo_entity = None
        self.previous = None
        self.clean()

    def clean(self, empty=False):
        self.previous = None
        self.cgo_entity = []
        if not empty:
            self.cgo_entity.extend(self.cgo_entity_begin)

    def new(self):
        self.previous = None

    def get(self):
        return self.cgo_entity + self.cgo_entity_end

    @staticmethod
    def make_color_triple(color_definition):
        return tuple(color_definition)


class BasicPymolCGOLines(BasicPymolCGO):
    cgo_entity_begin = [cgo.BEGIN, cgo.LINES]
    cgo_entity_end = [cgo.END]

    def add(self, coords=None, color=None):

        if color is not None:
            self.cgo_entity.append(cgo.COLOR)
            # self.cgo_entity.extend(map(float, color))
            self.cgo_entity.append(self.make_color_triple(list(map(float, color))))

        if coords is not None:
            for nr, coord in enumerate(coords):
                if self.previous is not None:
                    self.cgo_entity.append(cgo.VERTEX)
                    self.cgo_entity.extend(list(map(float, self.previous)))

                    self.cgo_entity.append(cgo.VERTEX)
                    self.cgo_entity.extend(list(map(float, coord)))
                self.previous = coord


class BasicPymolCGOSpheres(BasicPymolCGO):
    cgo_entity_begin = []
    cgo_entity_end = []

    def add(self, coords=None, radius=None, color=None):
        # color to colors...
        if color is not None:
            color = np.matrix(color).A
        if radius is not None:
            radius = np.matrix(radius).A1

        if coords is not None:
            for nr, coord in enumerate(coords):
                if color is not None:
                    if len(color) > 1:
                        c = color[nr]
                    else:
                        c = color[0]
                    self.cgo_entity.append(cgo.COLOR)
                    self.cgo_entity.append(self.make_color_triple(list(map(float, c))))
                    # self.cgo_entity.extend(map(float, c))
                self.cgo_entity.append(cgo.SPHERE)
                self.cgo_entity.extend(list(map(float, coord)))
                if radius is not None:
                    if len(radius) > 1:
                        r = radius[nr]
                    else:
                        r = radius
                else:
                    r = 1.
                self.cgo_entity.append(float(r))


# [cgo.CONE] + xyz3 + xyz2 + [hradius, 0.0] + color2 + color2

class BasicPymolCGOPointers(BasicPymolCGO):
    cgo_entity_begin = []
    cgo_entity_end = []

    def add_cone(self, coords1=None, coords2=None, radius1=None, radius2=None, color1=None, color2=None):
        # color to colors... ???
        if coords1 is not None and coords2 is not None:
            self.cgo_entity.append(cgo.CONE)
            self.cgo_entity.extend(list(map(float, coords1)))
            self.cgo_entity.extend(list(map(float, coords2)))
            self.cgo_entity.append(float(radius1))
            self.cgo_entity.append(float(radius2))
            self.cgo_entity.append(self.make_color_triple(list(map(float, color1))))
            self.cgo_entity.append(self.make_color_triple(list(map(float, color2))))
            # self.cgo_entity.extend(map(float, color1))
            # self.cgo_entity.extend(map(float, color2))

            self.cgo_entity.extend([cgo.NULL, cgo.POINTS])

    def add_pointer(self, point=None, direction=None, length=None, color=None, reverse=False):
        vec = point - direction
        vec_len = np.sqrt(np.sum(vec ** 2))
        vec = (vec / vec_len) * length
        vec = point + vec
        if reverse:
            self.add_cone(coords1=point, coords2=vec, radius1=length / 3., radius2=0, color1=color, color2=color)
        else:
            self.add_cone(coords1=vec, coords2=point, radius1=length / 3., radius2=0, color1=color, color2=color)


class SimpleTarWriteHelper(object):
    def __init__(self):
        self.tar_fh = None
        self.tmp_file = create_tmpfile()

    def open(self, filename):
        self.tar_fh = tarfile.open(filename, 'w:gz')

    def save_object2tar(self, obj, name):
        with open(self.tmp_file, 'wb') as f:
            pickle.dump(obj, f, protocol=2) # this is for compatibility with python2
            #json.dump(obj,f)
        self.save_file2tar(self.tmp_file, name)

    def save_file2tar(self, filename, name):
        self.tar_fh.add(filename, arcname=name)

    def __del__(self):
        if self.tar_fh is not None:
            self.tar_fh.close()
        os.unlink(self.tmp_file)


class ConnectToPymol(object):
    cgo_line_width = 2.
    ct_pymol = 'pymol'
    ct_file = 'file'

    def __init__(self):
        self.connection_type = None  # possible types are pymol and file

        self.script_fh = None
        self.data_fh = SimpleTarWriteHelper()

        self.cmd = None

    @listify
    def decode_color(self, cgo_object):
        for element in cgo_object:
            if isinstance(element, tuple):
                for e in element:
                    yield e
            else:
                yield element

    def init_pymol(self):
        import pymol
        self.cmd = pymol.cmd
        pymol.finish_launching()
        self.cmd.set('cgo_line_width', ConnectToPymol.cgo_line_width)
        self.connection_type = self.ct_pymol

    def init_script(self, filename):
        self.script_fh = open(filename, 'w')
        data_filename = os.path.splitext(os.path.basename(filename))[0] + '.tar.gz'
        self.data_fh.open(data_filename)
        self.connection_type = self.ct_file
        # TODO: Script generated by this object does not work properly if run via PyMOL.
        # init lines, imports etc.
        self.script_fh.write('''import argparse
parser=argparse.ArgumentParser(description="Aqua-Duct visualization script")
parser.add_argument("--save-session",action="store",dest="session",required=False,default=None,help="Pymol session file name.")
parser.add_argument("--discard",action="store",dest="discard",required=False,default='',help="Objects to discard.")
parser.add_argument("--keep",action="store",dest="keep",required=False,default='',help="Objects to keep.")
parser.add_argument("--force-color",action="store",dest="fc",required=False,default='',help="Force specific color.")
parser.add_argument("--fast",action="store_true",dest="fast",required=False,help="Hides all objects while loading.")
args,unknown=parser.parse_known_args()
import sys
if unknown: print("WARNING: Unknown options were used: "+" ".join(unknown), sys.stderr)
def _kd_order():
    if args.keep=='' and args.discard!='': return 'd'
    if args.keep!='' and args.discard=='': return 'k'
    if args.keep=='' and args.discard=='': return None
    if sys.argv.index('--keep')<sys.argv.index('--discard'): return 'k'
    return 'd'
kd_order = _kd_order()
def discard(name):
    if len([d for d in args.discard.split() if d in name])>0: return True
    return False
def keep(name):
    if len([k for k in args.keep.split() if k in name])>0: return True
    return False
def proceed(name):
    if kd_order == 'k':
        if not keep(name): return False
        elif discard(name): return False
    elif kd_order == 'd':
        if discard(name):
            if not keep(name): return False
    return True
from pymol import cmd,finish_launching
finish_launching()
print("Loading Aqua-Duct visualization...")
cmd.set("cgo_line_width",%d)
cmd.set("line_smooth","off")
from os import close,unlink
from os.path import splitext,isfile
import tarfile
import pickle as pickle
import json
from tempfile import mkstemp
fd, pdb_filename = mkstemp(suffix=".pdb")
close(fd)
max_state=0
arch_file="%s"
if not isfile(arch_file):
    import pymol
    if pymol.IS_WINDOWS:
        print("Please open visualization script using 'Open with' context menu and choose PyMol executable.")
        print("Alternatively, if you have PyMol installed as Python module, open visulaization script with Python executable.")
    while (pymol._ext_gui is None): pymol = reload(pymol)
    while (not hasattr(pymol._ext_gui,'root')): pymol = reload(pymol)
    import tkFileDialog
    arch_file=tkFileDialog.askopenfilename(filetypes=[("AQ Vis Arch","*.tar.gz")],title="Select AQ visualization archive",parent=pymol._ext_gui.root)
data_fh=tarfile.open(arch_file,"r:gz")
def decode_color(cgo_object,fc=None):
    for element in cgo_object:
        if isinstance(element,tuple):
            if fc is None:
                for e in element: yield e
            else:
                for e in fc: yield e
        else:
            yield element
def load_object(filename,name,state):
    if not proceed(name): return
    global max_state
    print("Loading %s" % splitext(filename)[0])
    obj=pickle.load(data_fh.extractfile(filename))
    #obj=json.load(data_fh.extractfile(filename))
    if name in args.fc.split():
        forced_color=args.fc.split()[args.fc.split().index(name)+1]
        forced_color=cmd.get_color_tuple(forced_color)
        obj=decode_color(obj,fc=forced_color)
    else:
        obj=decode_color(obj)
    cmd.load_cgo(obj,name,state)
    if state<2:
        if args.fast: cmd.disable("all")
        else: cmd.refresh()
    if state>max_state:
        max_state=state
def load_pdb(filename,name,state):
    if not proceed(name): return
    global max_state
    with open(pdb_filename,'wb') as fpdb:
        fpdb.write(data_fh.extractfile(filename).read())
    cmd.load(pdb_filename,state=state,object=name)
    if state>max_state:
        max_state=state
''' % (self.cgo_line_width, data_filename, "%s", "% s"))

    def add_cgo_object(self, name, cgo_object, state=None):
        if state is None:
            state = 1
        if self.connection_type == self.ct_pymol:
            self.cmd.load_cgo(self.decode_color(cgo_object), str(name), state)
        elif self.connection_type == self.ct_file:
            obj_name = '%s_%d.dump' % (name, state)
            self.data_fh.save_object2tar(cgo_object, obj_name)

            self.script_fh.write('''load_object("%s","%s",%d)''' % (obj_name, str(name), state))
            self.script_fh.write(os.linesep)
            # self.script_fh.write('''self.cmd.refresh()''')
            # self.script_fh.write(os.linesep)

    def del_cgo_object(self, name, state=None):
        raise NotImplementedError("Deletion of CGO objects is not implemented yet.")

    def load_pdb(self, name, filename, state=None):
        if state is None:
            state = 1
        if self.connection_type == self.ct_pymol:
            self.cmd.load(filename, state=state, object=name)
            self.cmd.show_as('cartoon', name)
            self.cmd.color('silver', name)
        elif self.connection_type == self.ct_file:
            # save pdblile as string
            filename_new = '%s_%d.pdb' % (name, state)
            self.data_fh.save_file2tar(filename, filename_new)
            self.script_fh.write('''load_pdb("%s","%s",%d)''' % (filename_new, name, state))
            self.script_fh.write(os.linesep)
            self.script_fh.write('''if proceed("%s"): cmd.show_as('cartoon','%s')''' % (name, name))
            self.script_fh.write(os.linesep)
            self.script_fh.write('''if proceed("%s"): cmd.color('silver','%s')''' % (name, name))
            self.script_fh.write(os.linesep)

    def orient_on(self, name):
        if self.connection_type == self.ct_pymol:
            self.cmd.orient(name)
        elif self.connection_type == self.ct_file:
            self.script_fh.write('''if proceed("%s"): cmd.orient("%s")''' % (name, name))
            self.script_fh.write(os.linesep)

    def __del__(self):
        if self.connection_type == self.ct_file:
            self.script_fh.write('''data_fh.close()
unlink(pdb_filename)
if args.fast: cmd.enable("all")
print("Aqua-Duct visualization loaded.")
if args.session:
    print("Preparing data to save session...")
    for state in range(max_state):
        cmd.set_frame(state+1)
        cmd.refresh()
        if (state+1)%100==0:
            print("wait... %d of %d done..." % (state+1,max_state))
    print("%d of %d done." % (state+1,max_state))
    print("Saving session...")
    cmd.set_frame(1)
    cmd.save(args.session,state=0)
    print("Let the Valve be always open!")
    print("Goodbye!")
    cmd.quit()
''')
            self.script_fh.write(os.linesep)
            self.script_fh.close()


class SinglePathPlotter(object):
    def __init__(self, pymol_connector, linearize=None):

        self.cgo_lines = BasicPymolCGOLines()
        self.cgo_spheres = BasicPymolCGOSpheres()
        self.cgo_pointers = BasicPymolCGOPointers()

        # linearize have to be callable! or False or None
        self.linearize = linearize

        self.pymol_connector = pymol_connector

    # TODO: take care of proper colors handling for smoothed and not smoothed traces!

    def add_single_path_continous_trace(self,
                                        spath,
                                        smooth=None,
                                        plot_in=True,
                                        plot_object=True,
                                        plot_out=True,
                                        plot_walk=True,
                                        **kwargs):

        self.cgo_lines.new()

        # get coords
        coords_cont = spath.get_coords_cont(smooth)

        # create slices
        if smooth:
            sls = tuple(list_blocks_to_slices(spath.types_cont))
        else:
            sls = tuple(list_blocks_to_slices(spath.etypes_cont))

        # create traces
        traces_list = tuple([coords_cont[sl] for sl in sls])
        new_line = False
        for trace, sl in zip(traces.midpoints(traces_list), sls):
            if len(trace) > 0:
                # now trace has midpoints
                # get type and etype
                t = spath.types_cont[sl][0]
                et = spath.etypes_cont[sl][0]
                if smooth:
                    et = et[0]
                # plot, if allowed
                if (plot_in and t == PathTypesCodes.path_in_code) or (
                        plot_object and t == PathTypesCodes.path_object_code) or (
                        plot_out and t == PathTypesCodes.path_out_code) or (
                        plot_walk and t == PathTypesCodes.path_walk_code):
                    # get color
                    c = color_codes(et)
                    # now, it is possible to linearize!
                    if smooth and self.linearize:
                        trace = self.linearize(trace)
                    self.cgo_lines.add(trace, cc(c))
                    # new_line = True

            if new_line:
                self.cgo_lines.new()
                new_line = False

    def paths_trace(self, spaths,
                    smooth=None,
                    name='paths',
                    state=None,
                    **kwargs):

        # if state_function is None
        if state is None:
            state = 1
        self.cgo_lines.clean()
        for nr, spath in enumerate(spaths):
            self.add_single_path_continous_trace(spath, smooth=smooth, **kwargs)
        if self.cgo_lines.cgo_entity == self.cgo_lines.cgo_entity_begin:
            self.cgo_lines.clean(empty=True)
        else:
            self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)

    def paths_inlets(self, spaths,
                     smooth=None,
                     color=None,
                     plot_in=True,
                     plot_out=True,
                     name='in-out-let',
                     state=None,
                     **kwargs):

        if state is None:
            state = 1
        self.cgo_pointers.clean()
        for nr, spath in enumerate(spaths):
            coords = spath.get_coords_cont(smooth=smooth)
            etypes = spath.etypes_cont
            if plot_in:
                if color is None:
                    c = color_codes(etypes[0])
                else:
                    c = color
                if len(coords) > 1:
                    self.cgo_pointers.add_pointer(point=coords[0],
                                                  direction=coords[1],
                                                  length=1.,
                                                  color=cc(c))
                else:
                    self.cgo_spheres.add(coords=coords, radius=None, color=cc(c))
            if plot_out:
                if color is None:
                    c = color_codes(etypes[-1])
                else:
                    c = color
                if len(coords) > 1:
                    self.cgo_pointers.add_pointer(point=coords[-1],
                                                  direction=coords[-2],
                                                  length=1.,
                                                  color=cc(c),
                                                  reverse=True)
                else:
                    self.cgo_spheres.add(coords=coords, radius=None, color=cc(c))
        self.pymol_connector.add_cgo_object(name, self.cgo_pointers.get(), state=state)

    def scatter(self, coords,
                radius=0.4,
                color='r',
                name='scatter',
                state=None):

        if isinstance(color, str):
            color = cc(color)
        if state is None:
            state = 1

        self.cgo_spheres.clean()

        self.cgo_spheres.add(coords=coords, radius=radius, color=color)

        self.pymol_connector.add_cgo_object(name, self.cgo_spheres.get(), state=state)

    def line(self, coords, color='r', name='line', state=None):

        if isinstance(color, str):
            color = cc(color)
        if state is None:
            state = 1

        self.cgo_lines.clean()

        self.cgo_lines.add(coords=coords, color=color)

        self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)

    def multiline_begin(self):
        self.cgo_lines.clean()

    def multiline_add(self, coords, color='r'):
        if isinstance(color, str):
            color = cc(color)
        self.cgo_lines.new()
        self.cgo_lines.add(coords=coords, color=color)

    def multiline_end(self, name=line, state=None):
        if state is None:
            state = 1
        self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)

    def convexhull(self, chull,
                   color='m',
                   name='convexhull',
                   state=None):

        if isinstance(color, str):
            color = cc(color)
        if state is None:
            state = 1

        self.cgo_lines.clean()
        for nr, facet in enumerate(chull.facets):
            if nr > 0:
                self.cgo_lines.new()
            self.cgo_lines.add(facet, color=color)

        self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)
